"""
This type stub file was generated by pyright.
"""

import operator
import numpy as np
import pytest
import pandas.util._test_decorators as td
import pandas as pd
import pandas._testing as tm
import zoneinfo
from collections import abc
from typing import Callable
from pandas.core.dtypes.dtypes import DatetimeTZDtype, IntervalDtype
from pandas import DataFrame, Interval, Period, Series, Timedelta, Timestamp
from pandas.core import ops
from pandas.core.indexes.api import Index, MultiIndex

"""
This file is very long and growing, but it was decided to not split it yet, as
it's still manageable (2020-03-17, ~1.1k LoC). See gh-31989

Instead of splitting it was decided to define sections here:
- Configuration / Settings
- Autouse fixtures
- Common arguments
- Missing values & co.
- Classes
- Indices
- Series'
- DataFrames
- Operators & Operations
- Data sets/files
- Time zones
- Dtypes
- Misc
"""
zoneinfo = ...
if pd.compat.PY39:
    ...
suppress_npdev_promotion_warning = ...
def pytest_addoption(parser) -> None:
    ...

def ignore_doctest_warning(item: pytest.Item, path: str, message: str) -> None:
    """Ignore doctest warning.

    Parameters
    ----------
    item : pytest.Item
        pytest test item.
    path : str
        Module path to Python object, e.g. "pandas.core.frame.DataFrame.append". A
        warning will be filtered when item.name ends with in given path. So it is
        sufficient to specify e.g. "DataFrame.append".
    message : str
        Message to be filtered.
    """
    ...

def pytest_collection_modifyitems(items, config): # -> None:
    ...

@pytest.fixture
def add_doctest_imports(doctest_namespace) -> None:
    """
    Make `np` and `pd` names available for doctests.
    """
    ...

@pytest.fixture(autouse=True)
def configure_tests() -> None:
    """
    Configure settings for all tests and test modules.
    """
    ...

@pytest.fixture(params=[0, 1, "index", "columns"], ids=lambda x: <Expression>)
def axis(request):
    """
    Fixture for returning the axis numbers of a DataFrame.
    """
    ...

axis_frame = ...
@pytest.fixture(params=[1, "columns"], ids=lambda x: <Expression>)
def axis_1(request):
    """
    Fixture for returning aliases of axis 1 of a DataFrame.
    """
    ...

@pytest.fixture(params=[True, False, None])
def observed(request):
    """
    Pass in the observed keyword to groupby for [True, False]
    This indicates whether categoricals should return values for
    values which are not in the grouper [False / None], or only values which
    appear in the grouper [True]. [None] is supported for future compatibility
    if we decide to change the default (and would need to warn if this
    parameter is not passed).
    """
    ...

@pytest.fixture(params=[True, False, None])
def ordered(request):
    """
    Boolean 'ordered' parameter for Categorical.
    """
    ...

@pytest.fixture(params=["first", "last", False])
def keep(request):
    """
    Valid values for the 'keep' parameter used in
    .duplicated or .drop_duplicates
    """
    ...

@pytest.fixture(params=["both", "neither", "left", "right"])
def inclusive_endpoints_fixture(request):
    """
    Fixture for trying all interval 'inclusive' parameters.
    """
    ...

@pytest.fixture(params=["left", "right", "both", "neither"])
def closed(request):
    """
    Fixture for trying all interval closed parameters.
    """
    ...

@pytest.fixture(params=["left", "right", "both", "neither"])
def other_closed(request):
    """
    Secondary closed fixture to allow parametrizing over all pairs of closed.
    """
    ...

@pytest.fixture(params=[None, "gzip", "bz2", "zip", "xz", "tar", pytest.param("zstd", marks=td.skip_if_no("zstandard"))])
def compression(request):
    """
    Fixture for trying common compression types in compression tests.
    """
    ...

@pytest.fixture(params=["gzip", "bz2", "zip", "xz", "tar", pytest.param("zstd", marks=td.skip_if_no("zstandard"))])
def compression_only(request):
    """
    Fixture for trying common compression types in compression tests excluding
    uncompressed case.
    """
    ...

@pytest.fixture(params=[True, False])
def writable(request):
    """
    Fixture that an array is writable.
    """
    ...

@pytest.fixture(params=["inner", "outer", "left", "right"])
def join_type(request):
    """
    Fixture for trying all types of join operations.
    """
    ...

@pytest.fixture(params=["nlargest", "nsmallest"])
def nselect_method(request):
    """
    Fixture for trying all nselect methods.
    """
    ...

@pytest.fixture(params=tm.NULL_OBJECTS, ids=lambda x: type(x).__name__)
def nulls_fixture(request):
    """
    Fixture for each null type in pandas.
    """
    ...

nulls_fixture2 = ...
@pytest.fixture(params=[None, np.nan, pd.NaT])
def unique_nulls_fixture(request):
    """
    Fixture for each null type in pandas, each null type exactly once.
    """
    ...

unique_nulls_fixture2 = ...
@pytest.fixture(params=tm.NP_NAT_OBJECTS, ids=lambda x: type(x).__name__)
def np_nat_fixture(request):
    """
    Fixture for each NaT type in numpy.
    """
    ...

np_nat_fixture2 = ...
@pytest.fixture(params=[DataFrame, Series])
def frame_or_series(request):
    """
    Fixture to parametrize over DataFrame and Series.
    """
    ...

@pytest.fixture(params=[Index, Series], ids=["index", "series"])
def index_or_series(request):
    """
    Fixture to parametrize over Index and Series, made necessary by a mypy
    bug, giving an error:

    List item 0 has incompatible type "Type[Series]"; expected "Type[PandasObject]"

    See GH#29725
    """
    ...

index_or_series2 = ...
@pytest.fixture(params=[Index, Series, pd.array], ids=["index", "series", "array"])
def index_or_series_or_array(request):
    """
    Fixture to parametrize over Index, Series, and ExtensionArray
    """
    ...

@pytest.fixture(params=[Index, Series, DataFrame, pd.array], ids=lambda x: x.__name__)
def box_with_array(request):
    """
    Fixture to test behavior for Index, Series, DataFrame, and pandas Array
    classes
    """
    ...

box_with_array2 = ...
@pytest.fixture
def dict_subclass(): # -> Type[TestSubDict]:
    """
    Fixture for a dictionary subclass.
    """
    class TestSubDict(dict):
        ...
    
    

@pytest.fixture
def non_dict_mapping_subclass(): # -> Type[TestNonDictMapping]:
    """
    Fixture for a non-mapping dictionary subclass.
    """
    class TestNonDictMapping(abc.Mapping):
        ...
    
    

@pytest.fixture
def multiindex_year_month_day_dataframe_random_data(): # -> Series* | Any | DataFrame | Series:
    """
    DataFrame with 3 level MultiIndex (year, month, day) covering
    first 100 business days from 2000-01-01 with random data
    """
    ...

@pytest.fixture
def lexsorted_two_level_string_multiindex() -> MultiIndex:
    """
    2-level MultiIndex, lexsorted, with string names.
    """
    ...

@pytest.fixture
def multiindex_dataframe_random_data(lexsorted_two_level_string_multiindex) -> DataFrame:
    """DataFrame with 2 level MultiIndex with random data"""
    ...

indices_dict = ...
if has_pyarrow:
    idx = ...
@pytest.fixture(params=indices_dict.keys())
def index(request):
    """
    Fixture for many "simple" kinds of indices.

    These indices are unlikely to cover corner cases, e.g.
        - no names
        - no NaTs/NaNs
        - no values near implementation bounds
        - ...
    """
    ...

index_fixture2 = ...
@pytest.fixture(params=[key for key in indices_dict if notisinstance(indices_dict[key], MultiIndex)])
def index_flat(request):
    """
    index fixture, but excluding MultiIndex cases.
    """
    ...

index_flat2 = ...
@pytest.fixture(params=[key for key in indices_dict if not(key in ["int", "uint", "range", "empty", "repeats", "bool-dtype"] or key.startswith("num_")) and notisinstance(indices_dict[key], MultiIndex)])
def index_with_missing(request): # -> MultiIndex | Any:
    """
    Fixture for indices with missing values.

    Integer-dtype and empty cases are excluded because they cannot hold missing
    values.

    MultiIndex is excluded because isna() is not defined for MultiIndex.
    """
    ...

@pytest.fixture
def string_series() -> Series:
    """
    Fixture for Series of floats with Index of unique strings
    """
    ...

@pytest.fixture
def object_series() -> Series:
    """
    Fixture for Series of dtype object with Index of unique strings
    """
    ...

@pytest.fixture
def datetime_series() -> Series:
    """
    Fixture for Series of floats with DatetimeIndex
    """
    ...

_series = ...
@pytest.fixture
def series_with_simple_index(index) -> Series:
    """
    Fixture for tests on series with changing types of indices.
    """
    ...

@pytest.fixture
def series_with_multilevel_index() -> Series:
    """
    Fixture with a Series with a 2-level MultiIndex.
    """
    ...

_narrow_series = ...
_index_or_series_objs = ...
@pytest.fixture(params=_index_or_series_objs.keys())
def index_or_series_obj(request):
    """
    Fixture for tests on indexes, series and series with a narrow dtype
    copy to avoid mutation, e.g. setting .name
    """
    ...

@pytest.fixture
def int_frame() -> DataFrame:
    """
    Fixture for DataFrame of ints with index of unique strings

    Columns are ['A', 'B', 'C', 'D']

                A  B  C  D
    vpBeWjM651  1  0  1  0
    5JyxmrP1En -1  0  0  0
    qEDaoD49U2 -1  1  0  0
    m66TkTfsFe  0  0  0  0
    EHPaNzEUFm -1  0 -1  0
    fpRJCevQhi  2  0  0  0
    OlQvnmfi3Q  0  0 -2  0
    ...        .. .. .. ..
    uB1FPlz4uP  0  0  0  1
    EcSe6yNzCU  0  0 -1  0
    L50VudaiI8 -1  1 -2  0
    y3bpw4nwIp  0 -1  0  0
    H0RdLLwrCT  1  1  0  0
    rY82K0vMwm  0  0  0  0
    1OPIUjnkjk  2  0  0  0

    [30 rows x 4 columns]
    """
    ...

@pytest.fixture
def datetime_frame() -> DataFrame:
    """
    Fixture for DataFrame of floats with DatetimeIndex

    Columns are ['A', 'B', 'C', 'D']

                       A         B         C         D
    2000-01-03 -1.122153  0.468535  0.122226  1.693711
    2000-01-04  0.189378  0.486100  0.007864 -1.216052
    2000-01-05  0.041401 -0.835752 -0.035279 -0.414357
    2000-01-06  0.430050  0.894352  0.090719  0.036939
    2000-01-07 -0.620982 -0.668211 -0.706153  1.466335
    2000-01-10 -0.752633  0.328434 -0.815325  0.699674
    2000-01-11 -2.236969  0.615737 -0.829076 -1.196106
    ...              ...       ...       ...       ...
    2000-02-03  1.642618 -0.579288  0.046005  1.385249
    2000-02-04 -0.544873 -1.160962 -0.284071 -1.418351
    2000-02-07 -2.656149 -0.601387  1.410148  0.444150
    2000-02-08 -1.201881 -1.289040  0.772992 -1.445300
    2000-02-09  1.377373  0.398619  1.008453 -0.928207
    2000-02-10  0.473194 -0.636677  0.984058  0.511519
    2000-02-11 -0.965556  0.408313 -1.312844 -0.381948

    [30 rows x 4 columns]
    """
    ...

@pytest.fixture
def float_frame() -> DataFrame:
    """
    Fixture for DataFrame of floats with index of unique strings

    Columns are ['A', 'B', 'C', 'D'].

                       A         B         C         D
    P7GACiRnxd -0.465578 -0.361863  0.886172 -0.053465
    qZKh6afn8n -0.466693 -0.373773  0.266873  1.673901
    tkp0r6Qble  0.148691 -0.059051  0.174817  1.598433
    wP70WOCtv8  0.133045 -0.581994 -0.992240  0.261651
    M2AeYQMnCz -1.207959 -0.185775  0.588206  0.563938
    QEPzyGDYDo -0.381843 -0.758281  0.502575 -0.565053
    r78Jwns6dn -0.653707  0.883127  0.682199  0.206159
    ...              ...       ...       ...       ...
    IHEGx9NO0T -0.277360  0.113021 -1.018314  0.196316
    lPMj8K27FA -1.313667 -0.604776 -1.305618 -0.863999
    qa66YMWQa5  1.110525  0.475310 -0.747865  0.032121
    yOa0ATsmcE -0.431457  0.067094  0.096567 -0.264962
    65znX3uRNG  1.528446  0.160416 -0.109635 -0.032987
    eCOBvKqf3e  0.235281  1.622222  0.781255  0.392871
    xSucinXxuV -1.263557  0.252799 -0.552247  0.400426

    [30 rows x 4 columns]
    """
    ...

@pytest.fixture
def mixed_type_frame() -> DataFrame:
    """
    Fixture for DataFrame of float/int/string columns with RangeIndex
    Columns are ['a', 'b', 'c', 'float32', 'int32'].
    """
    ...

@pytest.fixture
def rand_series_with_duplicate_datetimeindex() -> Series:
    """
    Fixture for Series with a DatetimeIndex that has duplicates.
    """
    ...

@pytest.fixture(params=[(Interval(left=0, right=5), IntervalDtype("int64", "right")), (Interval(left=0.1, right=0.5), IntervalDtype("float64", "right")), (Period("2012-01", freq="M"), "period[M]"), (Period("2012-02-01", freq="D"), "period[D]"), (Timestamp("2011-01-01", tz="US/Eastern"), DatetimeTZDtype(tz="US/Eastern")), (Timedelta(seconds=500), "timedelta64[ns]")])
def ea_scalar_and_dtype(request):
    ...

_all_arithmetic_operators = ...
@pytest.fixture(params=_all_arithmetic_operators)
def all_arithmetic_operators(request):
    """
    Fixture for dunder names for common arithmetic operations.
    """
    ...

@pytest.fixture(params=[operator.add, ops.radd, operator.sub, ops.rsub, operator.mul, ops.rmul, operator.truediv, ops.rtruediv, operator.floordiv, ops.rfloordiv, operator.mod, ops.rmod, operator.pow, ops.rpow, operator.eq, operator.ne, operator.lt, operator.le, operator.gt, operator.ge, operator.and_, ops.rand_, operator.xor, ops.rxor, operator.or_, ops.ror_])
def all_binary_operators(request):
    """
    Fixture for operator and roperator arithmetic, comparison, and logical ops.
    """
    ...

@pytest.fixture(params=[operator.add, ops.radd, operator.sub, ops.rsub, operator.mul, ops.rmul, operator.truediv, ops.rtruediv, operator.floordiv, ops.rfloordiv, operator.mod, ops.rmod, operator.pow, ops.rpow])
def all_arithmetic_functions(request):
    """
    Fixture for operator and roperator arithmetic functions.

    Notes
    -----
    This includes divmod and rdivmod, whereas all_arithmetic_operators
    does not.
    """
    ...

_all_numeric_reductions = ...
@pytest.fixture(params=_all_numeric_reductions)
def all_numeric_reductions(request):
    """
    Fixture for numeric reduction names.
    """
    ...

_all_boolean_reductions = ...
@pytest.fixture(params=_all_boolean_reductions)
def all_boolean_reductions(request):
    """
    Fixture for boolean reduction names.
    """
    ...

_all_reductions = ...
@pytest.fixture(params=_all_reductions)
def all_reductions(request):
    """
    Fixture for all (boolean + numeric) reduction names.
    """
    ...

@pytest.fixture(params=[operator.eq, operator.ne, operator.gt, operator.ge, operator.lt, operator.le])
def comparison_op(request):
    """
    Fixture for operator module comparison functions.
    """
    ...

@pytest.fixture(params=["__le__", "__lt__", "__ge__", "__gt__"])
def compare_operators_no_eq_ne(request):
    """
    Fixture for dunder names for compare operations except == and !=

    * >=
    * >
    * <
    * <=
    """
    ...

@pytest.fixture(params=["__and__", "__rand__", "__or__", "__ror__", "__xor__", "__rxor__"])
def all_logical_operators(request):
    """
    Fixture for dunder names for common logical operations

    * |
    * &
    * ^
    """
    ...

@pytest.fixture
def strict_data_files(pytestconfig):
    """
    Returns the configuration for the test setting `--strict-data-files`.
    """
    ...

@pytest.fixture
def datapath(strict_data_files: str) -> Callable[..., str]:
    """
    Get the path to a data file.

    Parameters
    ----------
    path : str
        Path to the file, relative to ``pandas/tests/``

    Returns
    -------
    path including ``pandas/tests``.

    Raises
    ------
    ValueError
        If the path doesn't exist and the --strict-data-files option is set.
    """
    ...

@pytest.fixture
def iris(datapath) -> DataFrame:
    """
    The iris dataset as a DataFrame.
    """
    ...

TIMEZONES = ...
if zoneinfo is not None:
    ...
TIMEZONE_IDS = ...
@td.parametrize_fixture_doc(str(TIMEZONE_IDS))
@pytest.fixture(params=TIMEZONES, ids=TIMEZONE_IDS)
def tz_naive_fixture(request):
    """
    Fixture for trying timezones including default (None): {0}
    """
    ...

@td.parametrize_fixture_doc(str(TIMEZONE_IDS[1]))
@pytest.fixture(params=TIMEZONES[1], ids=TIMEZONE_IDS[1])
def tz_aware_fixture(request):
    """
    Fixture for trying explicit timezones: {0}
    """
    ...

tz_aware_fixture2 = ...
_UTCS = ...
if zoneinfo is not None:
    ...
@pytest.fixture(params=_UTCS)
def utc_fixture(request):
    """
    Fixture to provide variants of UTC timezone strings and tzinfo objects.
    """
    ...

utc_fixture2 = ...
@pytest.fixture(params=tm.STRING_DTYPES)
def string_dtype(request):
    """
    Parametrized fixture for string dtypes.

    * str
    * 'str'
    * 'U'
    """
    ...

@pytest.fixture(params=["string[python]", pytest.param("string[pyarrow]", marks=td.skip_if_no("pyarrow", min_version="1.0.0"))])
def nullable_string_dtype(request):
    """
    Parametrized fixture for string dtypes.

    * 'string[python]'
    * 'string[pyarrow]'
    """
    ...

@pytest.fixture(params=["python", pytest.param("pyarrow", marks=td.skip_if_no("pyarrow", min_version="1.0.0"))])
def string_storage(request):
    """
    Parametrized fixture for pd.options.mode.string_storage.

    * 'python'
    * 'pyarrow'
    """
    ...

string_storage2 = ...
@pytest.fixture(params=tm.BYTES_DTYPES)
def bytes_dtype(request):
    """
    Parametrized fixture for bytes dtypes.

    * bytes
    * 'bytes'
    """
    ...

@pytest.fixture(params=tm.OBJECT_DTYPES)
def object_dtype(request):
    """
    Parametrized fixture for object dtypes.

    * object
    * 'object'
    """
    ...

@pytest.fixture(params=["object", "string[python]", pytest.param("string[pyarrow]", marks=td.skip_if_no("pyarrow", min_version="1.0.0"))])
def any_string_dtype(request):
    """
    Parametrized fixture for string dtypes.
    * 'object'
    * 'string[python]'
    * 'string[pyarrow]'
    """
    ...

@pytest.fixture(params=tm.DATETIME64_DTYPES)
def datetime64_dtype(request):
    """
    Parametrized fixture for datetime64 dtypes.

    * 'datetime64[ns]'
    * 'M8[ns]'
    """
    ...

@pytest.fixture(params=tm.TIMEDELTA64_DTYPES)
def timedelta64_dtype(request):
    """
    Parametrized fixture for timedelta64 dtypes.

    * 'timedelta64[ns]'
    * 'm8[ns]'
    """
    ...

@pytest.fixture
def fixed_now_ts() -> Timestamp:
    """
    Fixture emits fixed Timestamp.now()
    """
    ...

@pytest.fixture(params=tm.FLOAT_NUMPY_DTYPES)
def float_numpy_dtype(request):
    """
    Parameterized fixture for float dtypes.

    * float
    * 'float32'
    * 'float64'
    """
    ...

@pytest.fixture(params=tm.FLOAT_EA_DTYPES)
def float_ea_dtype(request):
    """
    Parameterized fixture for float dtypes.

    * 'Float32'
    * 'Float64'
    """
    ...

@pytest.fixture(params=tm.FLOAT_NUMPY_DTYPES + tm.FLOAT_EA_DTYPES)
def any_float_dtype(request):
    """
    Parameterized fixture for float dtypes.

    * float
    * 'float32'
    * 'float64'
    * 'Float32'
    * 'Float64'
    """
    ...

@pytest.fixture(params=tm.COMPLEX_DTYPES)
def complex_dtype(request):
    """
    Parameterized fixture for complex dtypes.

    * complex
    * 'complex64'
    * 'complex128'
    """
    ...

@pytest.fixture(params=tm.SIGNED_INT_NUMPY_DTYPES)
def any_signed_int_numpy_dtype(request):
    """
    Parameterized fixture for signed integer dtypes.

    * int
    * 'int8'
    * 'int16'
    * 'int32'
    * 'int64'
    """
    ...

@pytest.fixture(params=tm.UNSIGNED_INT_NUMPY_DTYPES)
def any_unsigned_int_numpy_dtype(request):
    """
    Parameterized fixture for unsigned integer dtypes.

    * 'uint8'
    * 'uint16'
    * 'uint32'
    * 'uint64'
    """
    ...

@pytest.fixture(params=tm.ALL_INT_NUMPY_DTYPES)
def any_int_numpy_dtype(request):
    """
    Parameterized fixture for any integer dtype.

    * int
    * 'int8'
    * 'uint8'
    * 'int16'
    * 'uint16'
    * 'int32'
    * 'uint32'
    * 'int64'
    * 'uint64'
    """
    ...

@pytest.fixture(params=tm.ALL_INT_EA_DTYPES)
def any_int_ea_dtype(request):
    """
    Parameterized fixture for any nullable integer dtype.

    * 'UInt8'
    * 'Int8'
    * 'UInt16'
    * 'Int16'
    * 'UInt32'
    * 'Int32'
    * 'UInt64'
    * 'Int64'
    """
    ...

@pytest.fixture(params=tm.ALL_INT_NUMPY_DTYPES + tm.ALL_INT_EA_DTYPES)
def any_int_dtype(request):
    """
    Parameterized fixture for any nullable integer dtype.

    * int
    * 'int8'
    * 'uint8'
    * 'int16'
    * 'uint16'
    * 'int32'
    * 'uint32'
    * 'int64'
    * 'uint64'
    * 'UInt8'
    * 'Int8'
    * 'UInt16'
    * 'Int16'
    * 'UInt32'
    * 'Int32'
    * 'UInt64'
    * 'Int64'
    """
    ...

@pytest.fixture(params=tm.ALL_INT_EA_DTYPES + tm.FLOAT_EA_DTYPES)
def any_numeric_ea_dtype(request):
    """
    Parameterized fixture for any nullable integer dtype and
    any float ea dtypes.

    * 'UInt8'
    * 'Int8'
    * 'UInt16'
    * 'Int16'
    * 'UInt32'
    * 'Int32'
    * 'UInt64'
    * 'Int64'
    * 'Float32'
    * 'Float64'
    """
    ...

@pytest.fixture(params=tm.SIGNED_INT_EA_DTYPES)
def any_signed_int_ea_dtype(request):
    """
    Parameterized fixture for any signed nullable integer dtype.

    * 'Int8'
    * 'Int16'
    * 'Int32'
    * 'Int64'
    """
    ...

@pytest.fixture(params=tm.ALL_REAL_NUMPY_DTYPES)
def any_real_numpy_dtype(request):
    """
    Parameterized fixture for any (purely) real numeric dtype.

    * int
    * 'int8'
    * 'uint8'
    * 'int16'
    * 'uint16'
    * 'int32'
    * 'uint32'
    * 'int64'
    * 'uint64'
    * float
    * 'float32'
    * 'float64'
    """
    ...

@pytest.fixture(params=tm.ALL_NUMPY_DTYPES)
def any_numpy_dtype(request):
    """
    Parameterized fixture for all numpy dtypes.

    * bool
    * 'bool'
    * int
    * 'int8'
    * 'uint8'
    * 'int16'
    * 'uint16'
    * 'int32'
    * 'uint32'
    * 'int64'
    * 'uint64'
    * float
    * 'float32'
    * 'float64'
    * complex
    * 'complex64'
    * 'complex128'
    * str
    * 'str'
    * 'U'
    * bytes
    * 'bytes'
    * 'datetime64[ns]'
    * 'M8[ns]'
    * 'timedelta64[ns]'
    * 'm8[ns]'
    * object
    * 'object'
    """
    ...

@pytest.fixture(params=tm.ALL_REAL_NUMPY_DTYPES + tm.COMPLEX_DTYPES + tm.ALL_INT_EA_DTYPES + tm.FLOAT_EA_DTYPES)
def any_numeric_dtype(request):
    """
    Parameterized fixture for all numeric dtypes.

    * int
    * 'int8'
    * 'uint8'
    * 'int16'
    * 'uint16'
    * 'int32'
    * 'uint32'
    * 'int64'
    * 'uint64'
    * float
    * 'float32'
    * 'float64'
    * complex
    * 'complex64'
    * 'complex128'
    * 'UInt8'
    * 'Int8'
    * 'UInt16'
    * 'Int16'
    * 'UInt32'
    * 'Int32'
    * 'UInt64'
    * 'Int64'
    * 'Float32'
    * 'Float64'
    """
    ...

_any_skipna_inferred_dtype = ...
@pytest.fixture(params=_any_skipna_inferred_dtype, ids=ids)
def any_skipna_inferred_dtype(request): # -> tuple[Unknown, NDArray[Any]]:
    """
    Fixture for all inferred dtypes from _libs.lib.infer_dtype

    The covered (inferred) types are:
    * 'string'
    * 'empty'
    * 'bytes'
    * 'mixed'
    * 'mixed-integer'
    * 'mixed-integer-float'
    * 'floating'
    * 'integer'
    * 'decimal'
    * 'boolean'
    * 'datetime64'
    * 'datetime'
    * 'date'
    * 'timedelta'
    * 'time'
    * 'period'
    * 'interval'

    Returns
    -------
    inferred_dtype : str
        The string for the inferred dtype from _libs.lib.infer_dtype
    values : np.ndarray
        An array of object dtype that will be inferred to have
        `inferred_dtype`

    Examples
    --------
    >>> import pandas._libs.lib as lib
    >>>
    >>> def test_something(any_skipna_inferred_dtype):
    ...     inferred_dtype, values = any_skipna_inferred_dtype
    ...     # will pass
    ...     assert lib.infer_dtype(values, skipna=True) == inferred_dtype
    """
    ...

@pytest.fixture
def ip():
    """
    Get an instance of IPython.InteractiveShell.

    Will raise a skip if IPython is not installed.
    """
    ...

@pytest.fixture(params=["bsr", "coo", "csc", "csr", "dia", "dok", "lil"])
def spmatrix(request): # -> Any:
    """
    Yields scipy sparse matrix classes.
    """
    ...

@pytest.fixture(params=[getattr(pd.offsets, o) for o in pd.offsets.__all__ if issubclass(getattr(pd.offsets, o), pd.offsets.Tick) and o != "Tick"])
def tick_classes(request):
    """
    Fixture for Tick based datetime offsets available for a time series.
    """
    ...

@pytest.fixture(params=[None, lambda x: x])
def sort_by_key(request):
    """
    Simple fixture for testing keys in sorting methods.
    Tests None (no key) and the identity key.
    """
    ...

@pytest.fixture()
def fsspectest(): # -> Generator[TestMemoryFS, None, None]:
    class TestMemoryFS(MemoryFileSystem):
        ...
    
    

@pytest.fixture(params=[("foo", None, None), ("Egon", "Venkman", None), ("NCC1701D", "NCC1701D", "NCC1701D"), (np.nan, np.nan, np.nan), (np.nan, pd.NaT, None), (np.nan, pd.NA, None), (pd.NA, pd.NA, pd.NA)])
def names(request):
    """
    A 3-tuple of names, the first two for operands, the last for a result.
    """
    ...

@pytest.fixture(params=[tm.setitem, tm.loc, tm.iloc])
def indexer_sli(request):
    """
    Parametrize over __setitem__, loc.__setitem__, iloc.__setitem__
    """
    ...

@pytest.fixture(params=[tm.loc, tm.iloc])
def indexer_li(request):
    """
    Parametrize over loc.__getitem__, iloc.__getitem__
    """
    ...

@pytest.fixture(params=[tm.setitem, tm.iloc])
def indexer_si(request):
    """
    Parametrize over __setitem__, iloc.__setitem__
    """
    ...

@pytest.fixture(params=[tm.setitem, tm.loc])
def indexer_sl(request):
    """
    Parametrize over __setitem__, loc.__setitem__
    """
    ...

@pytest.fixture(params=[tm.at, tm.loc])
def indexer_al(request):
    """
    Parametrize over at.__setitem__, loc.__setitem__
    """
    ...

@pytest.fixture(params=[tm.iat, tm.iloc])
def indexer_ial(request):
    """
    Parametrize over iat.__setitem__, iloc.__setitem__
    """
    ...

@pytest.fixture
def using_array_manager(): # -> bool | Any:
    """
    Fixture to check if the array manager is being used.
    """
    ...

@pytest.fixture
def using_copy_on_write() -> bool:
    """
    Fixture to check if Copy-on-Write is enabled.
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from textwrap import dedent
from typing import TYPE_CHECKING
from pandas._typing import Axis, TimedeltaConvertibleTypes
from pandas import DataFrame, Series
from pandas.core.generic import NDFrame
from pandas.util._decorators import doc
from pandas.core.window.doc import _shared_docs, args_compat, create_section_header, kwargs_compat, kwargs_numeric_only, numba_notes, template_header, template_returns, template_see_also, window_agg_numba_parameters
from pandas.core.window.rolling import BaseWindow, BaseWindowGroupby

if TYPE_CHECKING:
    ...
def get_center_of_mass(comass: float | None, span: float | None, halflife: float | None, alpha: float | None) -> float:
    ...

class ExponentialMovingWindow(BaseWindow):
    r"""
    Provide exponentially weighted (EW) calculations.

    Exactly one of ``com``, ``span``, ``halflife``, or ``alpha`` must be
    provided if ``times`` is not provided. If ``times`` is provided,
    ``halflife`` and one of ``com``, ``span`` or ``alpha`` may be provided.

    Parameters
    ----------
    com : float, optional
        Specify decay in terms of center of mass

        :math:`\alpha = 1 / (1 + com)`, for :math:`com \geq 0`.

    span : float, optional
        Specify decay in terms of span

        :math:`\alpha = 2 / (span + 1)`, for :math:`span \geq 1`.

    halflife : float, str, timedelta, optional
        Specify decay in terms of half-life

        :math:`\alpha = 1 - \exp\left(-\ln(2) / halflife\right)`, for
        :math:`halflife > 0`.

        If ``times`` is specified, a timedelta convertible unit over which an
        observation decays to half its value. Only applicable to ``mean()``,
        and halflife value will not apply to the other functions.

        .. versionadded:: 1.1.0

    alpha : float, optional
        Specify smoothing factor :math:`\alpha` directly

        :math:`0 < \alpha \leq 1`.

    min_periods : int, default 0
        Minimum number of observations in window required to have a value;
        otherwise, result is ``np.nan``.

    adjust : bool, default True
        Divide by decaying adjustment factor in beginning periods to account
        for imbalance in relative weightings (viewing EWMA as a moving average).

        - When ``adjust=True`` (default), the EW function is calculated using weights
          :math:`w_i = (1 - \alpha)^i`. For example, the EW moving average of the series
          [:math:`x_0, x_1, ..., x_t`] would be:

        .. math::
            y_t = \frac{x_t + (1 - \alpha)x_{t-1} + (1 - \alpha)^2 x_{t-2} + ... + (1 -
            \alpha)^t x_0}{1 + (1 - \alpha) + (1 - \alpha)^2 + ... + (1 - \alpha)^t}

        - When ``adjust=False``, the exponentially weighted function is calculated
          recursively:

        .. math::
            \begin{split}
                y_0 &= x_0\\
                y_t &= (1 - \alpha) y_{t-1} + \alpha x_t,
            \end{split}
    ignore_na : bool, default False
        Ignore missing values when calculating weights.

        - When ``ignore_na=False`` (default), weights are based on absolute positions.
          For example, the weights of :math:`x_0` and :math:`x_2` used in calculating
          the final weighted average of [:math:`x_0`, None, :math:`x_2`] are
          :math:`(1-\alpha)^2` and :math:`1` if ``adjust=True``, and
          :math:`(1-\alpha)^2` and :math:`\alpha` if ``adjust=False``.

        - When ``ignore_na=True``, weights are based
          on relative positions. For example, the weights of :math:`x_0` and :math:`x_2`
          used in calculating the final weighted average of
          [:math:`x_0`, None, :math:`x_2`] are :math:`1-\alpha` and :math:`1` if
          ``adjust=True``, and :math:`1-\alpha` and :math:`\alpha` if ``adjust=False``.

    axis : {0, 1}, default 0
        If ``0`` or ``'index'``, calculate across the rows.

        If ``1`` or ``'columns'``, calculate across the columns.

        For `Series` this parameter is unused and defaults to 0.

    times : str, np.ndarray, Series, default None

        .. versionadded:: 1.1.0

        Only applicable to ``mean()``.

        Times corresponding to the observations. Must be monotonically increasing and
        ``datetime64[ns]`` dtype.

        If 1-D array like, a sequence with the same shape as the observations.

        .. deprecated:: 1.4.0
            If str, the name of the column in the DataFrame representing the times.

    method : str {'single', 'table'}, default 'single'
        .. versionadded:: 1.4.0

        Execute the rolling operation per single column or row (``'single'``)
        or over the entire object (``'table'``).

        This argument is only implemented when specifying ``engine='numba'``
        in the method call.

        Only applicable to ``mean()``

    Returns
    -------
    ``ExponentialMovingWindow`` subclass

    See Also
    --------
    rolling : Provides rolling window calculations.
    expanding : Provides expanding transformations.

    Notes
    -----
    See :ref:`Windowing Operations <window.exponentially_weighted>`
    for further usage details and examples.

    Examples
    --------
    >>> df = pd.DataFrame({'B': [0, 1, 2, np.nan, 4]})
    >>> df
         B
    0  0.0
    1  1.0
    2  2.0
    3  NaN
    4  4.0

    >>> df.ewm(com=0.5).mean()
              B
    0  0.000000
    1  0.750000
    2  1.615385
    3  1.615385
    4  3.670213
    >>> df.ewm(alpha=2 / 3).mean()
              B
    0  0.000000
    1  0.750000
    2  1.615385
    3  1.615385
    4  3.670213

    **adjust**

    >>> df.ewm(com=0.5, adjust=True).mean()
              B
    0  0.000000
    1  0.750000
    2  1.615385
    3  1.615385
    4  3.670213
    >>> df.ewm(com=0.5, adjust=False).mean()
              B
    0  0.000000
    1  0.666667
    2  1.555556
    3  1.555556
    4  3.650794

    **ignore_na**

    >>> df.ewm(com=0.5, ignore_na=True).mean()
              B
    0  0.000000
    1  0.750000
    2  1.615385
    3  1.615385
    4  3.225000
    >>> df.ewm(com=0.5, ignore_na=False).mean()
              B
    0  0.000000
    1  0.750000
    2  1.615385
    3  1.615385
    4  3.670213

    **times**

    Exponentially weighted mean with weights calculated with a timedelta ``halflife``
    relative to ``times``.

    >>> times = ['2020-01-01', '2020-01-03', '2020-01-10', '2020-01-15', '2020-01-17']
    >>> df.ewm(halflife='4 days', times=pd.DatetimeIndex(times)).mean()
              B
    0  0.000000
    1  0.585786
    2  1.523889
    3  1.523889
    4  3.233686
    """
    _attributes = ...
    def __init__(self, obj: NDFrame, com: float | None = ..., span: float | None = ..., halflife: float | TimedeltaConvertibleTypes | None = ..., alpha: float | None = ..., min_periods: int | None = ..., adjust: bool = ..., ignore_na: bool = ..., axis: Axis = ..., times: str | np.ndarray | NDFrame | None = ..., method: str = ..., *, selection=...) -> None:
        ...
    
    def online(self, engine=..., engine_kwargs=...) -> OnlineExponentialMovingWindow:
        """
        Return an ``OnlineExponentialMovingWindow`` object to calculate
        exponentially moving window aggregations in an online method.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        engine: str, default ``'numba'``
            Execution engine to calculate online aggregations.
            Applies to all supported aggregation methods.

        engine_kwargs : dict, default None
            Applies to all supported aggregation methods.

            * For ``'numba'`` engine, the engine can accept ``nopython``, ``nogil``
              and ``parallel`` dictionary keys. The values must either be ``True`` or
              ``False``. The default ``engine_kwargs`` for the ``'numba'`` engine is
              ``{{'nopython': True, 'nogil': False, 'parallel': False}}`` and will be
              applied to the function

        Returns
        -------
        OnlineExponentialMovingWindow
        """
        ...
    
    @doc(_shared_docs["aggregate"], see_also=dedent("""
        See Also
        --------
        pandas.DataFrame.rolling.aggregate
        """), examples=dedent("""
        Examples
        --------
        >>> df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6], "C": [7, 8, 9]})
        >>> df
           A  B  C
        0  1  4  7
        1  2  5  8
        2  3  6  9

        >>> df.ewm(alpha=0.5).mean()
                  A         B         C
        0  1.000000  4.000000  7.000000
        1  1.666667  4.666667  7.666667
        2  2.428571  5.428571  8.428571
        """), klass="Series/Dataframe", axis="")
    def aggregate(self, func, *args, **kwargs): # -> Any | DataFrame | Series:
        ...
    
    agg = ...
    @doc(template_header, create_section_header("Parameters"), kwargs_numeric_only, args_compat, window_agg_numba_parameters(), kwargs_compat, create_section_header("Returns"), template_returns, create_section_header("See Also"), template_see_also, create_section_header("Notes"), numba_notes.replace("\n", "", 1), window_method="ewm", aggregation_description="(exponential weighted moment) mean", agg_method="mean")
    def mean(self, numeric_only: bool = ..., *args, engine=..., engine_kwargs=..., **kwargs): # -> DataFrame | Series:
        ...
    
    @doc(template_header, create_section_header("Parameters"), kwargs_numeric_only, args_compat, window_agg_numba_parameters(), kwargs_compat, create_section_header("Returns"), template_returns, create_section_header("See Also"), template_see_also, create_section_header("Notes"), numba_notes.replace("\n", "", 1), window_method="ewm", aggregation_description="(exponential weighted moment) sum", agg_method="sum")
    def sum(self, numeric_only: bool = ..., *args, engine=..., engine_kwargs=..., **kwargs): # -> DataFrame | Series:
        ...
    
    @doc(template_header, create_section_header("Parameters"), dedent("""
        bias : bool, default False
            Use a standard estimation bias correction.
        """).replace("\n", "", 1), kwargs_numeric_only, args_compat, kwargs_compat, create_section_header("Returns"), template_returns, create_section_header("See Also"), template_see_also[: -1], window_method="ewm", aggregation_description="(exponential weighted moment) standard deviation", agg_method="std")
    def std(self, bias: bool = ..., numeric_only: bool = ..., *args, **kwargs): # -> Any:
        ...
    
    def vol(self, bias: bool = ..., *args, **kwargs): # -> Any:
        ...
    
    @doc(template_header, create_section_header("Parameters"), dedent("""
        bias : bool, default False
            Use a standard estimation bias correction.
        """).replace("\n", "", 1), kwargs_numeric_only, args_compat, kwargs_compat, create_section_header("Returns"), template_returns, create_section_header("See Also"), template_see_also[: -1], window_method="ewm", aggregation_description="(exponential weighted moment) variance", agg_method="var")
    def var(self, bias: bool = ..., numeric_only: bool = ..., *args, **kwargs): # -> DataFrame | Series:
        ...
    
    @doc(template_header, create_section_header("Parameters"), dedent("""
        other : Series or DataFrame , optional
            If not supplied then will default to self and produce pairwise
            output.
        pairwise : bool, default None
            If False then only matching columns between self and other will be
            used and the output will be a DataFrame.
            If True then all pairwise combinations will be calculated and the
            output will be a MultiIndex DataFrame in the case of DataFrame
            inputs. In the case of missing elements, only complete pairwise
            observations will be used.
        bias : bool, default False
            Use a standard estimation bias correction.
        """).replace("\n", "", 1), kwargs_numeric_only, kwargs_compat, create_section_header("Returns"), template_returns, create_section_header("See Also"), template_see_also[: -1], window_method="ewm", aggregation_description="(exponential weighted moment) sample covariance", agg_method="cov")
    def cov(self, other: DataFrame | Series | None = ..., pairwise: bool | None = ..., bias: bool = ..., numeric_only: bool = ..., **kwargs): # -> DataFrame | Series:
        ...
    
    @doc(template_header, create_section_header("Parameters"), dedent("""
        other : Series or DataFrame, optional
            If not supplied then will default to self and produce pairwise
            output.
        pairwise : bool, default None
            If False then only matching columns between self and other will be
            used and the output will be a DataFrame.
            If True then all pairwise combinations will be calculated and the
            output will be a MultiIndex DataFrame in the case of DataFrame
            inputs. In the case of missing elements, only complete pairwise
            observations will be used.
        """).replace("\n", "", 1), kwargs_numeric_only, kwargs_compat, create_section_header("Returns"), template_returns, create_section_header("See Also"), template_see_also[: -1], window_method="ewm", aggregation_description="(exponential weighted moment) sample correlation", agg_method="corr")
    def corr(self, other: DataFrame | Series | None = ..., pairwise: bool | None = ..., numeric_only: bool = ..., **kwargs): # -> DataFrame | Series:
        ...
    


class ExponentialMovingWindowGroupby(BaseWindowGroupby, ExponentialMovingWindow):
    """
    Provide an exponential moving window groupby implementation.
    """
    _attributes = ...
    def __init__(self, obj, *args, _grouper=..., **kwargs) -> None:
        ...
    


class OnlineExponentialMovingWindow(ExponentialMovingWindow):
    def __init__(self, obj: NDFrame, com: float | None = ..., span: float | None = ..., halflife: float | TimedeltaConvertibleTypes | None = ..., alpha: float | None = ..., min_periods: int | None = ..., adjust: bool = ..., ignore_na: bool = ..., axis: Axis = ..., times: str | np.ndarray | NDFrame | None = ..., engine: str = ..., engine_kwargs: dict[str, bool] | None = ..., *, selection=...) -> None:
        ...
    
    def reset(self) -> None:
        """
        Reset the state captured by `update` calls.
        """
        ...
    
    def aggregate(self, func, *args, **kwargs): # -> Type[NotImplementedError]:
        ...
    
    def std(self, bias: bool = ..., *args, **kwargs): # -> Type[NotImplementedError]:
        ...
    
    def corr(self, other: DataFrame | Series | None = ..., pairwise: bool | None = ..., numeric_only: bool = ..., **kwargs): # -> Type[NotImplementedError]:
        ...
    
    def cov(self, other: DataFrame | Series | None = ..., pairwise: bool | None = ..., bias: bool = ..., numeric_only: bool = ..., **kwargs): # -> Type[NotImplementedError]:
        ...
    
    def var(self, bias: bool = ..., *args, **kwargs): # -> Type[NotImplementedError]:
        ...
    
    def mean(self, *args, update=..., update_times=..., **kwargs): # -> Series*:
        """
        Calculate an online exponentially weighted mean.

        Parameters
        ----------
        update: DataFrame or Series, default None
            New values to continue calculating the
            exponentially weighted mean from the last values and weights.
            Values should be float64 dtype.

            ``update`` needs to be ``None`` the first time the
            exponentially weighted mean is calculated.

        update_times: Series or 1-D np.ndarray, default None
            New times to continue calculating the
            exponentially weighted mean from the last values and weights.
            If ``None``, values are assumed to be evenly spaced
            in time.
            This feature is currently unsupported.

        Returns
        -------
        DataFrame or Series

        Examples
        --------
        >>> df = pd.DataFrame({"a": range(5), "b": range(5, 10)})
        >>> online_ewm = df.head(2).ewm(0.5).online()
        >>> online_ewm.mean()
              a     b
        0  0.00  5.00
        1  0.75  5.75
        >>> online_ewm.mean(update=df.tail(3))
                  a         b
        2  1.615385  6.615385
        3  2.550000  7.550000
        4  3.520661  8.520661
        >>> online_ewm.reset()
        >>> online_ewm.mean()
              a     b
        0  0.00  5.00
        1  0.75  5.75
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

from pandas.util._decorators import doc

"""

accessor.py contains base classes for implementing accessor properties
that can be mixed into or pinned onto other pandas classes.

"""
class DirNamesMixin:
    _accessors: set[str] = ...
    _hidden_attrs: frozenset[str] = ...
    def __dir__(self) -> list[str]:
        """
        Provide method name lookup and completion.

        Notes
        -----
        Only provide 'public' methods.
        """
        ...
    


class PandasDelegate:
    """
    Abstract base class for delegating methods/properties.
    """
    ...


def delegate_names(delegate, accessors, typ: str, overwrite: bool = ...): # -> (cls: Unknown) -> Unknown:
    """
    Add delegated names to a class using a class decorator.  This provides
    an alternative usage to directly calling `_add_delegate_accessors`
    below a class definition.

    Parameters
    ----------
    delegate : object
        The class to get methods/properties & doc-strings.
    accessors : Sequence[str]
        List of accessor to add.
    typ : {'property', 'method'}
    overwrite : bool, default False
       Overwrite the method/property in the target class if it exists.

    Returns
    -------
    callable
        A class decorator.

    Examples
    --------
    @delegate_names(Categorical, ["categories", "ordered"], "property")
    class CategoricalAccessor(PandasDelegate):
        [...]
    """
    ...

class CachedAccessor:
    """
    Custom property-like object.

    A descriptor for caching accessors.

    Parameters
    ----------
    name : str
        Namespace that will be accessed under, e.g. ``df.foo``.
    accessor : cls
        Class with the extension methods.

    Notes
    -----
    For accessor, The class's __init__ method assumes that one of
    ``Series``, ``DataFrame`` or ``Index`` as the
    single argument ``data``.
    """
    def __init__(self, name: str, accessor) -> None:
        ...
    
    def __get__(self, obj, cls):
        ...
    


@doc(_register_accessor, klass="DataFrame")
def register_dataframe_accessor(name): # -> (accessor: Unknown) -> Unknown:
    ...

@doc(_register_accessor, klass="Series")
def register_series_accessor(name): # -> (accessor: Unknown) -> Unknown:
    ...

@doc(_register_accessor, klass="Index")
def register_index_accessor(name): # -> (accessor: Unknown) -> Unknown:
    ...




# BoARIO : The Adaptative Regional Input Output model in python.
# Copyright (C) 2022  Samuel Juhel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
import json
import pathlib
import typing
from typing import Optional
import numpy as np
from numpy.typing import ArrayLike

from boario import logger, warn_once
from boario.event import Event
from pymrio.core.mriosystem import IOSystem
from boario.utils.misc import lexico_reindex


__all__ = ["ARIOBaseModel","INV_THRESHOLD","VALUE_ADDED_NAMES","VA_idx", "lexico_reindex"]

INV_THRESHOLD = 0 #20 #temporal_units

VALUE_ADDED_NAMES = ['VA', 'Value Added', 'value added',
                        'factor inputs', 'factor_inputs', 'Factors Inputs',
                        'Satellite Accounts', 'satellite accounts', 'satellite_accounts',
                     'satellite']

VA_idx = np.array(['Taxes less subsidies on products purchased: Total',
       'Other net taxes on production',
       "Compensation of employees; wages, salaries, & employers' social contributions: Low-skilled",
       "Compensation of employees; wages, salaries, & employers' social contributions: Medium-skilled",
       "Compensation of employees; wages, salaries, & employers' social contributions: High-skilled",
       'Operating surplus: Consumption of fixed capital',
       'Operating surplus: Rents on land',
       'Operating surplus: Royalties on resources',
       'Operating surplus: Remaining net operating surplus'], dtype=object)

class ARIOBaseModel(object):
    r"""The core of an ARIO3 model.  Handles the different arrays containing the mrio tables.

    A ARIOBaseModel wrap all the data and functions used in the core of the most basic version of the ARIO
    model (based on Hallegatte2013 and Guan2020).

    Attributes
    ----------

    results_storage : pathlib.Path
                      The path where the results of the simulation are stored.
    regions : numpy.ndarray of str
              An array of the regions of the model.
    n_regions : int
                The number :math:`m` of regions.
    sectors : numpy.ndarray of str
              An array of the sectors of the model.
    n_sectors : int
                The number :math:`n` of sectors.
    fd_cat : numpy.ndarray of str
             An array of the final demand categories of the model (``["Final demand"]`` if there is only one)
    n_fd_cat : int
               The numbers of final demand categories.
    monetary_unit : int
                    monetary unit prefix (i.e. if the tables unit is 10^6 € instead of 1 €, it should be set to 10^6).
    temporal_units_by_step : int
                     The number of temporal_units between each step. (Current version of the model was not tested with values other than `1`).
    year_to_temporal_unit_factor : int
                               Kinda deprecated, should be equal to `temporal_units_by_step`.
    rebuild_tau : int
                  Rebuilding characteristic time :math:`\tau_{\textrm{REBUILD}}` (see :ref:`boario-math`).
    overprod_max : float
                   Maximum factor of overproduction :math:`\alpha^{\textrm{max}}` (default should be 1.25).
    overprod_tau : float
                   Characteristic time of overproduction :math:`\tau_{\alpha}` in ``temporal_units_by_step`` (default should be 365 days).
    overprod_base : float
                    Base value of overproduction factor :math:`\alpha^{b}` (Default to 1.0).
    inv_duration : numpy.ndarray of int
                   Array :math:`\mathbf{s}` of size :math:`n` (sectors), setting for each input the initial number of ``temporal_units_by_step`` of stock for the input. (see :ref:`boario-math`).
    restoration_tau : numpy.ndarray of int
                      Array of size :math:`n` setting for each inputs its characteristic restoration time :math:`\tau_{\textrm{INV}}` in ``temporal_units_by_step``. (see :ref:`boario-math`).
    Z_0 : numpy.ndarray of float
          2-dim square matrix array :math:`\ioz` of size :math:`(n \times m, n \times m)` representing the intermediate (transaction) matrix (see :ref:`boario-math-init`).
    Z_C : numpy.ndarray of float
          2-dim matrix array :math:`\ioz^{\sectorsset}` of size :math:`(n, n \times m)` representing the intermediate (transaction) matrix aggregated by inputs (see :ref:`here <boario-math-z-agg>`).
    Z_distrib : numpy.ndarray of float
                :math:`\ioz` normalised by :math:`\ioz^{\sectorsset}`, i.e. representing for each input the share of the total ordered transiting from an industry to another (see :ref:`here <boario-math-z-agg>`).
    Y_0 : numpy.ndarray of float
          2-dim array :math:`\ioy` of size :math:`(n \times m, m \times \text{number of final demand categories})` representing the final demand matrix.
    X_0 : numpy.ndarray of float
          Array :math:`\iox(0)` of size :math:`n \times m` representing the initial gross production.
    gdp_df : pandas.DataFrame
             Dataframe of the total GDP of each region of the model
    VA_0 : numpy.ndarray of float
           Array :math:`\iov` of size :math:`n \times m` representing the total value added for each sectors.
    tech_mat : numpy.ndarray
               2-dim array :math:`\ioa` of size :math:`(n \times m, n \times m)` representing the technical coefficients matrix
    overprod : numpy.ndarray
               Array of size :math:`n \times m` representing the overproduction coefficients vector :math:`\mathbf{\alpha}(t)`.
    Raises
    ------
    RuntimeError
        A RuntimeError can occur when data is inconsistent (negative stocks for
        instance)
    ValueError
    NotImplementedError
    """

    def __init__(self,
                 pym_mrio: IOSystem,
                 mrio_params: dict,
                 simulation_params: dict,
                 results_storage: pathlib.Path
                 ) -> None:

        logger.debug("Initiating new ARIOBaseModel instance")
        super().__init__()

        ################ Parameters variables #######################
        logger.info("IO system metadata :\n{}".format(str(pym_mrio.meta)))
        logger.info("Simulation parameters:\n{}".format(json.dumps(simulation_params, indent=4)))
        pym_mrio = lexico_reindex(pym_mrio)
        self.mrio_params : dict = mrio_params
        self.main_inv_dur : int = mrio_params['main_inv_dur']

        results_storage = results_storage.absolute()
        self.results_storage:pathlib.Path = results_storage
        logger.info("Results storage is: {}".format(self.results_storage))
        reg = pym_mrio.get_regions()
        reg = typing.cast("list[str]",reg)
        self.regions = np.array(sorted(reg))
        self.n_regions = len(reg)
        sec = pym_mrio.get_sectors()
        sec = typing.cast("list[str]",sec)
        self.sectors = np.array(sorted(sec))
        self.n_sectors = len(sec)
        try:
            self.fd_cat = np.array(sorted(list(pym_mrio.get_Y_categories()))) #type: ignore
            self.n_fd_cat = len(pym_mrio.get_Y_categories()) #type: ignore
        except KeyError:
            self.n_fd_cat = 1
            self.fd_cat = np.array(["Final demand"])
        except IndexError:
            self.n_fd_cat= 1
            self.fd_cat = np.array(["Final demand"])
        self.monetary_unit = mrio_params['monetary_unit']
        logger.info("Monetary unit from params is: %s", self.monetary_unit)
        if pym_mrio.unit is None:
            raise ValueError("Problem reading monetary unit from MRIO")
        logger.info("Monetary unit from loaded mrio is: %s", pym_mrio.unit.unit.unique()[0])
        #self.psi = simulation_params['psi_param']
        self.n_temporal_units_by_step = simulation_params['temporal_units_by_step']
        self.iotable_year_to_temporal_unit_factor = simulation_params['year_to_temporal_unit_factor'] # 365 for yearly IO tables
        if self.iotable_year_to_temporal_unit_factor != 365:
            logger.warning("iotable_to_daily_step_factor is not set to 365 (days). This should probably not be the case if the IO tables you use are on a yearly basis.")
        self.steply_factor =  self.n_temporal_units_by_step / self.iotable_year_to_temporal_unit_factor
        self.rebuild_tau = simulation_params['rebuild_tau']
        self.overprod_max = simulation_params['alpha_max']
        self.overprod_tau = self.n_temporal_units_by_step / simulation_params['alpha_tau']
        self.overprod_base = simulation_params['alpha_base']
        inv = mrio_params['inventories_dict']
        inventories = [ np.inf if inv[k]=='inf' else inv[k] for k in sorted(inv.keys())]
        self.inv_duration = np.array(inventories)  / self.n_temporal_units_by_step
        if (self.inv_duration <= 1).any() :
            logger.warning("At least one product has inventory duration lower than the numbers of temporal units in one step ({}), model will set it to 2 by default, but you should probably check this !".format(self.n_temporal_units_by_step))
            self.inv_duration[self.inv_duration <= 1] = 2
        #################################################################


        ######## INITIAL MRIO STATE (in step temporality) ###############
        if pym_mrio.Z is None:
            raise ValueError("Z attribute of given MRIO doesn't exist, this is a problem")
        if pym_mrio.Y is None:
            raise ValueError("Y attribute of given MRIO doesn't exist, this is a problem")
        if pym_mrio.x is None:
            raise ValueError("x attribute of given MRIO doesn't exist, this is a problem")
        self._matrix_id = np.eye(self.n_sectors)
        self._matrix_I_sum = np.tile(self._matrix_id, self.n_regions)
        self.Z_0 = pym_mrio.Z.to_numpy()
        self.Z_C = (self._matrix_I_sum @ self.Z_0)
        with np.errstate(divide='ignore',invalid='ignore'):
            self.Z_distrib = (np.divide(self.Z_0,(np.tile(self.Z_C, (self.n_regions, 1)))))
        self.Z_distrib = np.nan_to_num(self.Z_distrib)
        self.Z_0 = (pym_mrio.Z.to_numpy() * self.steply_factor)
        self.Y_0 = (pym_mrio.Y.to_numpy() * self.steply_factor)
        tmp = np.array(pym_mrio.x.T)
        self.X_0 = (tmp.flatten() * self.steply_factor)
        del tmp
        value_added = (pym_mrio.x.T - pym_mrio.Z.sum(axis=0))
        value_added = value_added.reindex(sorted(value_added.index), axis=0)
        value_added = value_added.reindex(sorted(value_added.columns), axis=1)
        value_added[value_added < 0] = 0.0
        self.gdp_df = value_added.groupby('region',axis=1).sum()
        self.VA_0 = (value_added.to_numpy().flatten())
        self.tech_mat = ((self._matrix_I_sum @ pym_mrio.A).to_numpy()) #type: ignore #to_numpy is not superfluous !
        kratio = mrio_params['capital_ratio_dict']
        kratio_ordered = [kratio[k] for k in sorted(kratio.keys())]
        self.kstock_ratio_to_VA = np.tile(np.array(kratio_ordered),self.n_regions)
        if value_added.ndim > 1:
            self.gdp_share_sector = (self.VA_0 / value_added.sum(axis=0).groupby('region').transform('sum').to_numpy())
        else:
            self.gdp_share_sector = (self.VA_0 / value_added.groupby('region').transform('sum').to_numpy())
        self.gdp_share_sector = self.gdp_share_sector.flatten()
        self.matrix_share_thresh = self.Z_C > np.tile(self.X_0, (self.n_sectors, 1)) * 0.00001 # [n_sectors, n_regions*n_sectors]
        #################################################################

        ####### SIMULATION VARIABLES ####################################
        self.overprod = np.full((self.n_regions * self.n_sectors), self.overprod_base, dtype=np.float64)
        with np.errstate(divide='ignore',invalid='ignore'):
            self.matrix_stock = ((np.tile(self.X_0, (self.n_sectors, 1)) * self.tech_mat) * self.inv_duration[:,np.newaxis])
        self.matrix_stock = np.nan_to_num(self.matrix_stock,nan=np.inf, posinf=np.inf)
        self.matrix_stock_0 = self.matrix_stock.copy()
        self.matrix_orders = self.Z_0.copy()
        self.production = self.X_0.copy()
        self.intmd_demand = self.Z_0.copy()
        self.final_demand = self.Y_0.copy()
        self.rebuilding_demand = None
        #self.rebuild_demand = np.zeros(shape = np.concatenate([self.Z_0,self.Y_0],axis=1).shape)
        # self.prod_max_toward_rebuilding = None
        self.kapital_lost = np.zeros(self.production.shape)
        self.order_type = simulation_params['order_type']

        self.prod_cap_delta = np.zeros(shape = self.X_0.shape)
        #################################################################

        ################## SIMULATION TRACKING VARIABLES ################
        self.in_shortage = False
        self.had_shortage = False
        self.production_evolution = np.memmap(results_storage/"iotable_XVA_record", dtype='float64', mode="w+", shape=(simulation_params['n_temporal_units_to_sim'], self.n_sectors*self.n_regions))
        self.production_evolution.fill(np.nan)
        self.production_cap_evolution = np.memmap(results_storage/"iotable_X_max_record", dtype='float64', mode="w+", shape=(simulation_params['n_temporal_units_to_sim'], self.n_sectors*self.n_regions))
        self.production_cap_evolution.fill(np.nan)
        self.final_demand_evolution = np.memmap(results_storage/"final_demand_record", dtype='float64', mode="w+", shape=(simulation_params['n_temporal_units_to_sim'], self.n_sectors*self.n_regions))
        self.final_demand_evolution.fill(np.nan)
        self.io_demand_evolution = np.memmap(results_storage/"io_demand_record", dtype='float64', mode="w+", shape=(simulation_params['n_temporal_units_to_sim'], self.n_sectors*self.n_regions))
        self.io_demand_evolution.fill(np.nan)

        self.rebuild_demand_evolution = np.memmap(results_storage/"rebuild_demand_record", dtype='float64', mode="w+", shape=(simulation_params['n_temporal_units_to_sim'], self.n_sectors*self.n_regions))
        self.rebuild_demand_evolution.fill(np.nan)
        self.overproduction_evolution = np.memmap(results_storage/"overprodvector_record", dtype='float64', mode="w+", shape=(simulation_params['n_temporal_units_to_sim'], self.n_sectors*self.n_regions))
        self.overproduction_evolution.fill(np.nan)
        self.final_demand_unmet_evolution = np.memmap(results_storage/"final_demand_unmet_record", dtype='float64', mode="w+", shape=(simulation_params['n_temporal_units_to_sim'], self.n_sectors*self.n_regions))
        self.final_demand_unmet_evolution.fill(np.nan)
        self.rebuild_production_evolution = np.memmap(results_storage/"rebuild_prod_record", dtype='float64', mode="w+", shape=(simulation_params['n_temporal_units_to_sim'], self.n_sectors*self.n_regions))
        self.rebuild_production_evolution.fill(np.nan)
        if simulation_params['register_stocks']:
            self.stocks_evolution = np.memmap(results_storage/"stocks_record", dtype='float64', mode="w+", shape=(simulation_params['n_temporal_units_to_sim'], self.n_sectors, self.n_sectors*self.n_regions))
            self.stocks_evolution.fill(np.nan)
        self.limiting_stocks_evolution = np.memmap(results_storage/"limiting_stocks_record", dtype='byte', mode="w+", shape=(simulation_params['n_temporal_units_to_sim'], self.n_sectors, self.n_sectors*self.n_regions))
        self.limiting_stocks_evolution.fill(-1)
        #############################################################################

        ### Internals
        self._prod_delta_type = None

        #### POST INIT ####
        ### Event Class Attribute setting
        Event.possible_regions = self.regions.copy()
        Event.regions_idx = np.arange(self.n_regions)
        Event.possible_sectors = self.sectors.copy()
        Event.sectors_idx = np.arange(self.n_sectors)
        Event.temporal_unit_range = simulation_params["n_temporal_units_to_sim"]
        Event.z_shape = self.Z_0.shape
        Event.y_shape = self.Y_0.shape
        Event.x_shape = self.X_0.shape
        Event.monetary_unit = self.monetary_unit
        Event.sectors_gva_shares = self.gdp_share_sector.copy()
        Event.Z_distrib = self.Z_distrib.copy()

        meta = pym_mrio.meta.metadata
        Event.mrio_name = meta["name"] + "_" + meta["description"] + "_" + meta["system"] + "_" + meta["version"]

        # initialize those (it's not very nice, but otherwise python complains)
        self._indus_rebuild_demand_tot = None
        self._house_rebuild_demand_tot = None
        self._indus_rebuild_demand = None
        self._house_rebuild_demand = None
        self._tot_rebuild_demand = None
        self._kapital_lost = None
        self._prod_cap_delta_kapital = None
        self._prod_cap_delta_arbitrary = None
        self._prod_cap_delta_tot = None


        ### Dump indexes
        self.write_index(results_storage/"indexes.json")

    ## Properties

    @property
    def tot_rebuild_demand(self) -> Optional[np.ndarray]:
        tmp = []
        if self._indus_rebuild_demand_tot is not None:
            tmp.append(self._indus_rebuild_demand_tot)
        if self._house_rebuild_demand_tot is not None:
            tmp.append(self._house_rebuild_demand_tot)
        if tmp != []:
            ret = np.concatenate(tmp,axis=1).sum(axis=1)
            self._tot_rebuild_demand = ret
        else:
            self._tot_rebuild_demand = None
        return self._tot_rebuild_demand

    @tot_rebuild_demand.setter
    def tot_rebuild_demand(self, source:list[Event]):
        """Compute and update total rebuilding demand.

        Compute and update total rebuilding demand for the given list of events. Only events
        tagged as rebuildable are accounted for.

        TODO: ADD MATH

        Parameters
        ----------
        events : 'list[Event]'
            A list of Event objects
        """
        if not isinstance(source, list):
            ValueError("Setting tot_rebuild_demand can only be done with a list of events, not a {}".format(type(source)))

        self.house_rebuild_demand = source
        self.indus_rebuild_demand = source

    @property
    def house_rebuild_demand(self)->Optional[np.ndarray]:
        return self._house_rebuild_demand

    @property
    def house_rebuild_demand_tot(self)->Optional[np.ndarray]:
        return self._house_rebuild_demand_tot

    @house_rebuild_demand.setter
    def house_rebuild_demand(self, source:list[Event]):
        r"""Compute rebuild demand from household demand from a list of events

        Compute and return rebuilding final households demand for the given list of events
        by summing the rebuilding_demand_house member of each event. Only events
        tagged as rebuildable are accounted for. The shape of the array
        returned is the same as the final demand member (``Y_0`` | :math:`\ioy`) of the calling MrioSystem.

        Parameters
        ----------
        events : 'list[Event]'
            A list of Event objects

        Returns
        -------
        np.ndarray
            An array of same shape as Y_0, containing the sum of all currently
            rebuildable final demand stock from all events in the given list.

        Notes
        -----

        Currently the model wasn't tested with such a rebuilding demand. Only intermediate demand is considered.
        """
        tmp = []
        for ev in source:
            if ev.rebuildable:
                if ev.rebuild_tau is None:
                    rebuild_tau = self.rebuild_tau
                else:
                    rebuild_tau = ev.rebuild_tau
                    warn_once(logger, "Event has a custom rebuild_tau")
                if ev._rebuilding_demand_house is not None:
                    tmp.append(ev._rebuilding_demand_house * (self.n_temporal_units_by_step / rebuild_tau))
        if tmp == []:
            self._house_rebuild_demand = None
            self._house_rebuild_demand_tot = None
        else:
            house_reb_dem = np.stack(tmp,axis=-1)
            tot = house_reb_dem.sum(axis=1)
            self._house_rebuild_demand = house_reb_dem
            self._house_rebuild_demand_tot = tot

    @property
    def indus_rebuild_demand(self)->Optional[np.ndarray]:
        return self._indus_rebuild_demand

    @property
    def indus_rebuild_demand_tot(self)->Optional[np.ndarray]:
        return self._indus_rebuild_demand_tot

    @indus_rebuild_demand.setter
    def indus_rebuild_demand(self, source:list[Event]):
        r"""Compute rebuild demand from economic sectors (i.e. not households)

        Compute and return rebuilding 'intermediate demand' for the given list of events
        by summing the rebuildind_demand_indus of each event. Only events
        tagged as rebuildable are accounted for. The shape of the array
        returned is the same as the intermediate demand member (``Z_0`` | :math:`\ioz`) of the calling
        MrioSystem.

        Parameters
        ----------
        events : 'list[Event]'
            A list of Event objects

        Returns
        -------
        np.ndarray
            An array of same shape as Z_0, containing the sum of all currently
            rebuildable intermediate demand stock from all events in the given list.
        """
        tmp = []
        for ev in source:
            if ev.rebuildable:
                if ev.rebuild_tau is None:
                    rebuild_tau = self.rebuild_tau
                else:
                    rebuild_tau = ev.rebuild_tau
                    warn_once(logger, "Event has a custom rebuild_tau")
                tmp.append(ev._rebuilding_demand_indus * (self.n_temporal_units_by_step / rebuild_tau))

        if tmp == []:
            self._indus_rebuild_demand = None
            self._indus_rebuild_demand_tot = None
        else:
            indus_reb_dem = np.stack(tmp,axis=-1)
            self._indus_rebuild_demand = indus_reb_dem
            self._indus_rebuild_demand_tot = indus_reb_dem.sum(axis=1)
    @property
    def kapital_lost(self) -> Optional[np.ndarray]:
        return self._kapital_lost

    @kapital_lost.setter
    def kapital_lost(self, source:list[Event]|np.ndarray
                              ) -> None:
        if isinstance(source,list):
            if source != []:
                events_K = [ev for ev in source if "kapital_destroyed" in ev.shock_type]
                self._kapital_lost = np.add.reduce(np.array([e.regional_sectoral_kapital_destroyed for e in events_K]))
            else:
                self._kapital_lost = np.zeros(self.VA_0.shape)
        elif isinstance(source,np.ndarray):
            self._kapital_lost = source

        productivity_loss_from_K = np.zeros(shape=self._kapital_lost.shape)
        k_stock = (self.VA_0 * self.kstock_ratio_to_VA)
        np.divide(self._kapital_lost, k_stock, out=productivity_loss_from_K, where=k_stock!=0)
        self._prod_cap_delta_kapital = productivity_loss_from_K
        if (self._prod_cap_delta_kapital > 0.0).any():
            if self._prod_delta_type is None:
                self._prod_delta_type = "from_kapital"
            elif self._prod_delta_type == "from_arbitrary":
                self._prod_delta_type = "mixed_from_kapital_from_arbitrary"

    @property
    def prod_cap_delta_arbitrary(self) -> Optional[np.ndarray]:
        return self._prod_cap_delta_arbitrary

    @prod_cap_delta_arbitrary.setter
    def prod_cap_delta_arbitrary(self, source: list[Event]|np.ndarray):
        if isinstance(source,list):
            event_arb = np.array([ev for ev in source if ev.prod_cap_delta_arbitrary is not None])
            if event_arb.size == 0:
                self._prod_cap_delta_arbitrary = np.zeros(shape = self.X_0.shape)
            else:
                self._prod_cap_delta_arbitrary = np.max.reduce(event_arb)
        else:
            self._prod_capt_delta_arbitrary= source
        assert self._prod_cap_delta_arbitrary is not None
        if (self._prod_cap_delta_arbitrary > 0.0).any():
            if self._prod_delta_type is None:
                self._prod_delta_type = "from_arbitrary"
            elif self._prod_delta_type == "from_kapital":
                self._prod_delta_type = "mixed_from_kapital_from_arbitrary"

    @property
    def prod_cap_delta_kapital(self) -> Optional[np.ndarray]:
        return self._prod_cap_delta_kapital

    @property
    def prod_cap_delta_tot(self) -> np.ndarray:
        tmp = []
        if self._prod_delta_type is None:
            raise AttributeError("Production delta doesn't appear to be set yet.")
        elif self._prod_delta_type == "from_kapital":
            tmp.append(self._prod_cap_delta_kapital)
        elif self._prod_delta_type == "from_arbitrary":
            tmp.append(self._prod_cap_delta_arbitrary)
        elif self._prod_delta_type == "mixed_from_kapital_from_arbitrary":
            tmp.append(self._prod_cap_delta_kapital)
            tmp.append(self._prod_cap_delta_arbitrary)
        else:
            raise NotImplementedError("Production delta type {} not recognised".format(self._prod_delta_type))
        tmp.append(np.ones(shape=self.X_0.shape))
        #logger.debug("tmp: {}".format(tmp))
        self._prod_cap_delta_tot = np.amin(np.stack(tmp,axis=-1),axis=1)
        assert self._prod_cap_delta_tot.shape == self.X_0.shape, "expected shape {}, received {}".format(self.X_0.shape, self._prod_cap_delta_tot.shape)
        return self._prod_cap_delta_tot

    @prod_cap_delta_tot.setter
    def prod_cap_delta_tot(self, source: list[Event]):
        if not isinstance(source,list):
            ValueError("Setting prod_cap_delta_tot can only be done with a list of events, not a {}".format(type(source)))

        self.kapital_lost = source
        self.prod_cap_delta_arbitrary = source

    def update_system_from_events(self, events: list[Event]) -> None:
        """Update MrioSystem variables according to given list of events

        Compute kapital loss for each industry affected as the sum of their total rebuilding demand (to each rebuilding sectors and for each events). This information is stored as a 1D array ``kapital_lost`` of size :math:`n \time m`.
        Also compute the total rebuilding demand.

        Parameters
        ----------
        events : 'list[Event]'
            List of events (as Event objects) to consider.

        """
        self.prod_cap_delta_tot = events
        self.tot_rebuild_demand = events


    @property
    def production_cap(self)->np.ndarray:
        r"""Compute and update production capacity.

        Compute and update production capacity from possible kapital damage and overproduction.

        .. math::

            x^{Cap}_{f}(t) = \alpha_{f}(t) (1 - \Delta_{f}(t)) x_{f}(t)

        Raises
        ------
        ValueError
            Raised if any industry has negative production (probably from kapital loss too high)
        """
        production_cap = self.X_0.copy()
        if (self._prod_delta_type is not None):
            prod_delta_tot = self.prod_cap_delta_tot.copy()
            if (prod_delta_tot > 0.).any():
                production_cap = production_cap * (1 - prod_delta_tot)
        if (self.overprod > 1.0).any():
            production_cap = production_cap * self.overprod
        if (production_cap < 0).any() :
            raise ValueError("Production capacity was found negative for at least on industry")
        return production_cap

    @property
    def total_demand(self) -> np.ndarray:
        """Computes and updates total demand

        Update total rebuild demand (and apply rebuilding characteristic time
        if ``separate_rebuilding`` is False, then sum/reduce rebuilding, orders
        and final demand together and accordingly update vector of total
        demand.

        Parameters
        ----------
        events : 'list[Event]'
            List of Event to consider for rebuilding demand.
        separate_rebuilding : bool
            A boolean specifying if rebuilding demand should be treated as a whole (true) or under the characteristic time/proportional scheme strategy.
        Returns
        -------
        None

        """
        if (self.matrix_orders < 0).any() :
            raise RuntimeError("Some matrix orders are negative which shouldn't happen")

        tot_dem = self.matrix_orders.sum(axis=1) + self.final_demand.sum(axis=1)
        if (tot_dem < 0).any() :
            raise RuntimeError("Some total demand are negative which shouldn't happen")
        if self.tot_rebuild_demand is not None:
            tot_dem += self.tot_rebuild_demand
        return tot_dem

    @property
    def production_opt(self) -> np.ndarray:
        return np.fmin(self.total_demand, self.production_cap)

    @property
    def inventory_constraints_opt(self) -> np.ndarray:
        return self.calc_inventory_constraints(self.production_opt)

    @property
    def inventory_constraints_act(self) -> np.ndarray:
        return self.calc_inventory_constraints(self.production)

    def calc_production(self, current_temporal_unit:int) -> np.ndarray:
        r"""Compute and update actual production

        1. Compute ``production_opt`` and ``inventory_constraints`` as :

        .. math::
           :nowrap:

                \begin{alignat*}{4}
                      \iox^{\textrm{Opt}}(t) &= (x^{\textrm{Opt}}_{f}(t))_{f \in \firmsset} &&= \left ( \min \left ( d^{\textrm{Tot}}_{f}(t), x^{\textrm{Cap}}_{f}(t) \right ) \right )_{f \in \firmsset} && \text{Optimal production}\\
                      \mathbf{\ioinv}^{\textrm{Cons}}(t) &= (\omega^{\textrm{Cons},f}_p(t))_{\substack{p \in \sectorsset\\f \in \firmsset}} &&=
                    \begin{bmatrix}
                        \tau^{1}_1 & \hdots & \tau^{p}_1 \\
                        \vdots & \ddots & \vdots\\
                        \tau^1_n & \hdots & \tau^{p}_n
                    \end{bmatrix}
                    \odot \begin{bmatrix} \iox^{\textrm{Opt}}(t)\\ \vdots\\ \iox^{\textrm{Opt}}(t) \end{bmatrix} \odot \ioa^{\sectorsset} && \text{Inventory constraints} \\
                    &&&= \begin{bmatrix}
                        \tau^{1}_1 x^{\textrm{Opt}}_{1}(t) a_{11} & \hdots & \tau^{p}_1 x^{\textrm{Opt}}_{p}(t) a_{1p}\\
                        \vdots & \ddots & \vdots\\
                        \tau^1_n x^{\textrm{Opt}}_{1}(t) a_{n1} & \hdots & \tau^{p}_n x^{\textrm{Opt}}_{p}(t) a_{np}
                    \end{bmatrix} && \\
                \end{alignat*}

        2. If stocks do not meet ``inventory_constraints`` for any inputs, then decrease production accordingly :

        .. math::
           :nowrap:

                \begin{alignat*}{4}
                    \iox^{a}(t) &= (x^{a}_{f}(t))_{f \in \firmsset} &&= \left \{ \begin{aligned}
                                                           & x^{\textrm{Opt}}_{f}(t) & \text{if $\omega_{p}^f(t) \geq \omega^{\textrm{Cons},f}_p(t)$}\\
                                                           & x^{\textrm{Opt}}_{f}(t) \cdot \min_{p \in \sectorsset} \left ( \frac{\omega^s_{p}(t)}{\omega^{\textrm{Cons,f}}_p(t)} \right ) & \text{if $\omega_{p}^f(t) < \omega^{\textrm{Cons},f}_p(t)$}
                                                           \end{aligned} \right. \quad &&
                \end{alignat*}

        Also warns in log if such shortages happen.


        Parameters
        ----------
        current_temporal_unit : int
            current step number

        """
        #1.
        production_opt = self.production_opt.copy()
        inventory_constraints = self.inventory_constraints_opt.copy()
        #2.
        if (stock_constraint := (self.matrix_stock < inventory_constraints) * self.matrix_share_thresh).any():
            if not self.in_shortage:
                logger.info('At least one industry entered shortage regime. (step:{})'.format(current_temporal_unit))
            self.in_shortage = True
            self.had_shortage = True
            production_ratio_stock = np.ones(shape=self.matrix_stock.shape)
            np.divide(self.matrix_stock, inventory_constraints, out=production_ratio_stock, where=(self.matrix_share_thresh * (inventory_constraints!=0)))
            production_ratio_stock[production_ratio_stock > 1] = 1
            if (production_ratio_stock < 1).any():
                production_max = np.tile(production_opt, (self.n_sectors, 1)) * production_ratio_stock
                assert not (np.min(production_max,axis=0) < 0).any()
                self.production = np.min(production_max, axis=0)
            else:
                assert not (production_opt < 0).any()
                self.production = production_opt
        else:
            if self.in_shortage:
                self.in_shortage = False
                logger.info('All industries exited shortage regime. (step:{})'.format(current_temporal_unit))
            assert not (production_opt < 0).any()
            self.production = production_opt
        return stock_constraint

    def calc_inventory_constraints(self, production:np.ndarray) -> np.ndarray :
        """Compute inventory constraint (no psi parameter)

        See :meth:`calc_production`.

        Parameters
        ----------
        production : np.ndarray
            production vector

        Returns
        -------
        np.ndarray
            A boolean NDArray `stock_constraint` of the same shape as ``matrix_stock`` (ie `(n_sectors,n_regions*n_sectors)`), with ``True`` for any input not meeting the inventory constraints.

        """
        inventory_constraints = (np.tile(production, (self.n_sectors, 1)) * self.tech_mat)
        tmp = np.tile(np.nan_to_num(self.inv_duration, posinf=0.)[:,np.newaxis],(1,self.n_regions*self.n_sectors))
        return inventory_constraints * tmp



    def distribute_production(self,
                              current_temporal_unit: int, rebuildable_events: 'list[Event]',
                              scheme:str='proportional', separate_rebuilding:bool=False) -> list[Event]:
        r"""Production distribution module

    #. Computes rebuilding demand for each rebuildable events (applying the `rebuild_tau` characteristic time)

    #. Creates/Computes total demand matrix (Intermediate + Final + Rebuild)

    #. Assesses if total demand is greater than realized production, hence requiring rationning

    #. Distributes production proportionally to demand such that :

        .. math::
           :nowrap:

               \begin{alignat*}{4}
                   &\ioorders^{\textrm{Received}}(t) &&= \left (\frac{o_{ff'}(t)}{d^{\textrm{Tot}}_f(t)} \cdot x^a_f(t) \right )_{f,f'\in \firmsset}\\
                   &\ioy^{\textrm{Received}}(t) &&= \left ( \frac{y_{f,c}}{d^{\textrm{Tot}}_f(t)}\cdot x^a_f(t) \right )_{f\in \firmsset, c \in \catfdset}\\
                   &\Damage^{\textrm{Repaired}}(t) &&= \left ( \frac{\gamma_{f,c}}{d^{\textrm{Tot}}_f(t)} \cdot x^a_f(t) \right )_{f\in \firmsset, c \in \catfdset}\\
               \end{alignat*}

        Where :

        - :math:`\ioorders^{\textrm{Received}}(t)` is the received orders matrix,

        - :math:`\ioy^{\textrm{Received}}(t)` is the final demand received matrix,

        - :math:`\Damage^{\textrm{Repared}}(t)` is the rebuilding/repair achieved matrix,

        - :math:`d^{\textrm{Tot}}_f(t)` is the total demand to industry :math:`f`,

        - :math:`x^a_f(t)` is :math:`f`'s realized production,

        - :math:`o_{ff'}(t)` is the quantity of product ordered by industry :math:`f'` to industry :math:`f`,

        - :math:`y_{fc}(t)` is the quantity of product ordered by household :math:`c` to industry :math:`f`,

        - :math:`\gamma_{fc}(t)` is the repaired/rebuilding demand ordered to :math:`f`.

    #. Updates stocks matrix. (Only if `np.allclose(stock_add, stock_use).all()` is false)

        .. math::
           :nowrap:

               \begin{alignat*}{4}
                   &\ioinv(t+1) &&= \ioinv(t) + \left ( \mathbf{I}_{\textrm{sum}} \cdot \ioorders^{\textrm{Received}}(t) \right ) - \left ( \colvec{\iox^{\textrm{a}}(t)}{\iox^{\textrm{a}}(t)} \odot \ioa^{\sectorsset} \right )\\
               \end{alignat*}

        Where :

        - :math:`\ioinv` is the inventory matrix,
        - :math:`\mathbf{I}_{\textrm{sum}}` is a row summation matrix,
        - :math:`\ioa^{\sectorsset}` is the (input not specific to region) technical coefficients matrix.

    #. Computes final demand not met due to rationing and write it.

    #. Updates rebuilding demand for each event (by substracting distributed production)

    Parameters
    ----------
    current_temporal_unit : int
        Current temporal unit (day|week|... depending on parameters) (required to write the final demand not met)
    events : 'list[Event]'
        List of rebuildable events
    scheme : str
        Placeholder for future distribution scheme
    separate_rebuilding : bool
        Currently unused.

    Returns
    -------
    list[Event]
        The list of events to remove from current events (as they are totally rebuilt)

    Raises
    ------
    RuntimeError
        If negative values are found in places there's should not be any
    ValueError
        If an attempt to run an unimplemented distribution scheme is tried

"""

        if scheme != 'proportional':
            raise ValueError("Scheme %s not implemented"% scheme)

        list_of_demands = [self.matrix_orders, self.final_demand]
        ## 1. Calc demand from rebuilding requirements (with characteristic time rebuild_tau)
        if rebuildable_events != []:
            n_events = len(rebuildable_events)
            tot_rebuilding_demand_summed = self.tot_rebuild_demand.copy()
            # debugging assert
            assert tot_rebuilding_demand_summed.shape == self.X_0.shape
            indus_reb_dem_tot_per_event = self.indus_rebuild_demand_tot.copy()
            indus_reb_dem_per_event = self.indus_rebuild_demand.copy()

            # expected shape assert (debug also)
            exp_shape_indus_per_event = (self.n_sectors*self.n_regions,self.n_sectors*self.n_regions,n_events)
            exp_shape_indus_tot_per_event = (self.n_sectors*self.n_regions,n_events)
            assert indus_reb_dem_per_event.shape == exp_shape_indus_per_event, "expected shape is {}, given shape is {}".format(exp_shape_indus_per_event, indus_reb_dem_per_event.shape)
            assert indus_reb_dem_tot_per_event.shape == exp_shape_indus_tot_per_event, "expected shape is {}, given shape is {}".format(exp_shape_indus_tot_per_event, indus_reb_dem_tot_per_event.shape)

            house_reb_dem_tot_per_event = self.house_rebuild_demand_tot.copy()
            house_reb_dem_per_event = self.house_rebuild_demand.copy()

            # expected shape assert (debug also)
            exp_shape_house = (self.n_sectors*self.n_regions,self.n_fd_cat*self.n_regions,n_events)
            exp_shape_house_tot = (self.n_sectors*self.n_regions,n_events)
            assert house_reb_dem_per_event.shape == exp_shape_house
            assert house_reb_dem_tot_per_event.shape == exp_shape_house_tot
            h_reb = (house_reb_dem_tot_per_event > 0).any()
            ind_reb=(indus_reb_dem_tot_per_event > 0).any()
        else:
            tot_rebuilding_demand_summed = np.zeros(shape=self.X_0.shape)
            h_reb = False
            ind_reb = False

        ## 2. Concat to have total demand matrix (Intermediate + Final + Rebuild)
        tot_demand = np.concatenate([self.matrix_orders, self.final_demand, np.expand_dims(tot_rebuilding_demand_summed,1)], axis=1)
        ## 3. Does production meet total demand
        rationning_required = (self.production - tot_demand.sum(axis=1))<(-1/self.monetary_unit)
        rationning_mask = np.tile(rationning_required[:,np.newaxis],(1,tot_demand.shape[1]))
        demand_share = np.full(tot_demand.shape,0.0)
        tot_dem_summed = np.expand_dims(np.sum(tot_demand, axis=1, where=rationning_mask),1)
        # Get demand share
        np.divide(tot_demand, tot_dem_summed, where=(tot_dem_summed!=0), out=demand_share)
        distributed_production = tot_demand.copy()
        # 4. distribute production proportionally to demand
        np.multiply(demand_share, np.expand_dims(self.production,1), out=distributed_production, where=rationning_mask)
        intmd_distribution = distributed_production[:,:self.n_sectors * self.n_regions]
        # Stock use is simply production times technical coefs
        stock_use = np.tile(self.production, (self.n_sectors,1)) * self.tech_mat
        if (stock_use < 0).any() :
            raise RuntimeError("Stock use contains negative values, this should not happen")
        # 5. Restock is the production from each supplier, summed.
        stock_add = self._matrix_I_sum @ intmd_distribution
        if (stock_add < 0).any():
            raise RuntimeError("stock_add (restocking) contains negative values, this should not happen")
        if not np.allclose(stock_add, stock_use):
            self.matrix_stock = self.matrix_stock - stock_use + stock_add
            if (self.matrix_stock < 0).any():
                self.matrix_stock.dump(self.results_storage/"matrix_stock_dump.pkl")
                logger.error("Negative values in the stocks, matrix has been dumped in the results dir : \n {}".format(self.results_storage/"matrix_stock_dump.pkl"))
                raise RuntimeError("stock_add (restocking) contains negative values, this should not happen")

        # 6. Compute final demand not met due to rationing
        final_demand_not_met = self.final_demand - distributed_production[:,self.n_sectors*self.n_regions:(self.n_sectors*self.n_regions + self.n_fd_cat*self.n_regions)]
        final_demand_not_met = final_demand_not_met.sum(axis=1)
        # avoid -0.0 (just in case)
        final_demand_not_met[final_demand_not_met==0.] = 0.
        self.write_final_demand_unmet(current_temporal_unit, final_demand_not_met)

        # 7. Compute production delivered to rebuilding
        rebuild_prod = distributed_production[:,(self.n_sectors*self.n_regions + self.n_fd_cat*self.n_regions):].copy().flatten()
        self.write_rebuild_prod(current_temporal_unit,rebuild_prod) #type: ignore

        if h_reb and ind_reb:
            #logger.debug("Entering here")
            indus_shares = np.divide(indus_reb_dem_tot_per_event, tot_rebuilding_demand_summed, where=(tot_rebuilding_demand_summed!=0))
            house_shares = np.divide(house_reb_dem_tot_per_event, tot_rebuilding_demand_summed, where=(tot_rebuilding_demand_summed!=0))
            #logger.debug("indus_shares: {}".format(indus_shares))
            assert np.allclose(indus_shares + house_shares,np.ones(shape=indus_shares.shape))
        elif h_reb:
            house_shares = np.ones(house_reb_dem_tot_per_event.shape)
            indus_shares = np.zeros(indus_reb_dem_tot_per_event.shape)
        elif ind_reb:
            house_shares = np.zeros(house_reb_dem_tot_per_event.shape)
            indus_shares = np.ones(indus_reb_dem_tot_per_event.shape)
            #logger.debug("indus_shares: {}".format(indus_shares.shape))
        else:
            return []

        indus_rebuild_prod = rebuild_prod[:,np.newaxis] * indus_shares #type:ignore
        house_rebuild_prod = rebuild_prod[:,np.newaxis] * house_shares #type:ignore

        indus_rebuild_prod_distributed = np.zeros(shape=indus_reb_dem_per_event.shape)
        house_rebuild_prod_distributed = np.zeros(shape=house_reb_dem_per_event.shape)
        if ind_reb:
            # 1. We normalize rebuilding demand by total rebuilding demand (i.e. we get for each client asking, the share of the total demand)
            # This is done by broadcasting total demand and then dividing. (Perhaps this is not efficient ?)
            indus_rebuilding_demand_shares = np.zeros(shape=indus_reb_dem_per_event.shape)
            indus_rebuilding_demand_broad = np.broadcast_to(indus_reb_dem_tot_per_event[:,np.newaxis],indus_reb_dem_per_event.shape)
            np.divide(indus_reb_dem_per_event,indus_rebuilding_demand_broad, where=(indus_rebuilding_demand_broad!=0), out=indus_rebuilding_demand_shares)

            # 2. Then we multiply those shares by the total production (each client get production proportional to its demand relative to total demand)
            indus_rebuild_prod_broad = np.broadcast_to(indus_rebuild_prod[:,np.newaxis],indus_reb_dem_per_event.shape)
            np.multiply(indus_rebuilding_demand_shares,indus_rebuild_prod_broad,out=indus_rebuild_prod_distributed)

        if h_reb:
            house_rebuilding_demand_shares = np.zeros(shape=indus_reb_dem_per_event.shape)
            house_rebuilding_demand_broad = np.broadcast_to(house_reb_dem_tot_per_event[:,np.newaxis],house_reb_dem_per_event.shape)
            np.divide(house_reb_dem_per_event,house_rebuilding_demand_broad, where=(house_rebuilding_demand_broad!=0), out=house_rebuilding_demand_shares)
            house_rebuild_prod_broad = np.broadcast_to(house_rebuild_prod[:,np.newaxis],house_reb_dem_per_event.shape)
            np.multiply(house_rebuilding_demand_shares,house_rebuild_prod_broad,out=house_rebuild_prod_distributed)

        # update rebuilding demand
        events_to_remove = []
        for e_id, e in enumerate(rebuildable_events):
            if e.rebuilding_demand_indus is not None:
                e.rebuilding_demand_indus -= indus_rebuild_prod_distributed[:,:,e_id]
            if e.rebuilding_demand_house is not None:
                e.rebuilding_demand_house -= house_rebuild_prod_distributed[:,:,e_id]
            if (e.rebuilding_demand_indus < (10/self.monetary_unit)).all() and (e.rebuilding_demand_house < (10/self.monetary_unit)).all():
                events_to_remove.append(e)
        return events_to_remove

    def calc_matrix_stock_gap(self, matrix_stock_goal) -> np.ndarray:
        matrix_stock_gap = np.zeros(matrix_stock_goal.shape)
        # logger.debug("matrix_stock_goal: {}".format(matrix_stock_goal.shape))
        # logger.debug("matrix_stock: {}".format(self.matrix_stock.shape))
        # logger.debug("matrix_stock_goal_finite: {}".format(matrix_stock_goal[np.isfinite(matrix_stock_goal)].shape))
        # logger.debug("matrix_stock_finite: {}".format(self.matrix_stock[np.isfinite(self.matrix_stock)].shape))
        matrix_stock_gap[np.isfinite(matrix_stock_goal)] = (matrix_stock_goal[np.isfinite(matrix_stock_goal)] - self.matrix_stock[np.isfinite(self.matrix_stock)])
        if (np.isnan(matrix_stock_gap).any()):
            raise RuntimeError("NaN in matrix stock gap")
        matrix_stock_gap[matrix_stock_gap < 0] = 0
        return matrix_stock_gap


    def calc_orders(self) -> None:
        """TODO describe function

        :param stocks_constraints:
        :type stocks_constraints:
        :returns:

        """
        #total_demand = self.total_demand
        production_opt = self.production_opt.copy()
        matrix_stock_goal = np.tile(production_opt, (self.n_sectors, 1)) * self.tech_mat
        # Check this !
        with np.errstate(invalid='ignore'):
            matrix_stock_goal *= self.inv_duration[:,np.newaxis]
        if np.allclose(self.matrix_stock, matrix_stock_goal):
            matrix_stock_gap = matrix_stock_goal * 0
        else:
            matrix_stock_gap = self.calc_matrix_stock_gap(matrix_stock_goal)
        matrix_stock_gap += (np.tile(self.production, (self.n_sectors, 1)) * self.tech_mat)
        if self.order_type == "alt":
            prod_ratio = np.ones(shape=self.X_0.shape)
            np.divide(self.production,self.X_0, out=prod_ratio, where=self.X_0!=0)
            Z_prod = self.Z_0 * prod_ratio[:, np.newaxis]
            Z_Cprod = np.tile(self._matrix_I_sum @ Z_prod,(self.n_regions,1))
            out=np.zeros(shape=Z_prod.shape)
            np.divide(Z_prod,Z_Cprod,out=out, where=Z_Cprod!=0)
            tmp = (np.tile(matrix_stock_gap, (self.n_regions, 1)) * out)
        else:
            tmp = (np.tile(matrix_stock_gap, (self.n_regions, 1)) * self.Z_distrib)
        if (tmp < 0).any():
            raise RuntimeError("Negative orders computed")
        self.matrix_orders = tmp


    def calc_overproduction(self) -> None:
        scarcity = np.full(self.production.shape, 0.0)
        total_demand = self.total_demand.copy()
        scarcity[total_demand!=0] = (total_demand[total_demand!=0] - self.production[total_demand!=0]) / total_demand[total_demand!=0]
        scarcity[np.isnan(scarcity)] = 0
        overprod_chg = (((self.overprod_max - self.overprod) * (scarcity) * self.overprod_tau) + ((self.overprod_base - self.overprod) * (scarcity == 0) * self.overprod_tau)).flatten()
        self.overprod += overprod_chg
        self.overprod[self.overprod < 1.] = 1.

    def check_stock_increasing(self, current_temporal_unit:int):
        tmp = np.full(self.matrix_stock.shape,0.0)
        mask = np.isfinite(self.matrix_stock_0)
        np.subtract(self.matrix_stock,self.matrix_stock_0, out=tmp, where=mask)
        check_1 = tmp > 0.0
        tmp = np.full(self.matrix_stock.shape,0.0)
        np.subtract(self.stocks_evolution[current_temporal_unit], self.stocks_evolution[current_temporal_unit-1], out=tmp, where=mask)
        check_2 = (tmp >= 0.0)
        return (check_1 & check_2).all()

    def check_production_eq_strict(self):
        return ((np.isclose(self.production, self.X_0)) | np.greater(self.production, self.X_0)).all()

    def check_production_eq_soft(self, current_temporal_unit:int, period:int = 10) -> bool:
        return self.check_monotony(self.production_evolution, current_temporal_unit, period)

    def check_stocks_monotony(self, current_temporal_unit:int, period:int = 10) -> bool:
        return self.check_monotony(self.stocks_evolution, current_temporal_unit, period)

    def check_initial_equilibrium(self)-> bool:
        return (np.allclose(self.production, self.X_0) and np.allclose(self.matrix_stock, self.matrix_stock_0))

    def check_equilibrium_soft(self, current_temporal_unit:int):
        return (self.check_stock_increasing(current_temporal_unit) and self.check_production_eq_strict)

    def check_equilibrium_monotony(self, current_temporal_unit:int, period:int=10) -> bool:
        return self.check_production_eq_soft(current_temporal_unit, period) and self.check_stocks_monotony(current_temporal_unit, period)

    def check_monotony(self, x, current_temporal_unit:int, period:int = 10) -> bool:
        return np.allclose(x[current_temporal_unit], x[current_temporal_unit-period], atol=0.0001)

    def check_crash(self, prod_threshold : float=0.80) -> int:
        """Check for economic crash

        This method look at the production vector and returns the number of
        industries which production is less than a certain share (default 20%) of the starting
        production.

        Parameters
        ----------
        prod_threshold : float, default: 0.8
            An industry is counted as 'crashed' if its current production is less than its starting production times (1 - `prod_threshold`).


        """
        tmp = np.full(self.production.shape, 0.0)
        checker = np.full(self.production.shape, 0.0)
        mask = self.X_0 != 0
        np.subtract(self.X_0, self.production, out=tmp, where=mask)
        np.divide(tmp, self.X_0, out=checker, where=mask)
        return np.where(checker >= prod_threshold)[0].size

    def reset_module(self,
                 simulation_params: dict,
                 ) -> None:
        # Reset OUTPUTS
        logger.warning("This method is quite probably deprecated")
        self.reset_record_files(simulation_params['n_temporal_units_to_sim'], simulation_params['register_stocks'])
        # Reset variable attributes
        self.kapital_lost = np.zeros(self.production.shape)
        self.overprod = np.full((self.n_regions * self.n_sectors), self.overprod_base, dtype=np.float64)
        with np.errstate(divide='ignore',invalid='ignore'):
            self.matrix_stock = ((np.tile(self.X_0, (self.n_sectors, 1)) * self.tech_mat) * self.inv_duration[:,np.newaxis])
        self.matrix_stock = np.nan_to_num(self.matrix_stock,nan=np.inf, posinf=np.inf)
        self.matrix_stock_0 = self.matrix_stock.copy()
        self.matrix_orders = self.Z_0.copy()
        self.production = self.X_0.copy()
        self.intmd_demand = self.Z_0.copy()
        self.final_demand = self.Y_0.copy()
        self.rebuilding_demand = None
        self.prod_max_toward_rebuilding = None

    def update_params(self, new_params:dict) -> None:
        """Update the parameters of the model.

        Replace each parameters with given new ones.

        .. warning::
            Be aware this method calls :meth:`~boario.model_base.reset_record_files`, which resets the memmap files located in the results directory !

        Parameters
        ----------
        new_params : dict
            Dictionary of new parameters to use.

        """
        logger.warning("This method is quite probably deprecated")
        self.n_temporal_units_by_step = new_params['temporal_units_by_step']
        self.iotable_year_to_temporal_unit_factor = new_params['year_to_temporal_unit_factor']
        self.rebuild_tau = new_params['rebuild_tau']
        self.overprod_max = new_params['alpha_max']
        self.overprod_tau = new_params['alpha_tau']
        self.overprod_base = new_params['alpha_base']
        if self.results_storage != pathlib.Path(new_params['output_dir']+"/"+new_params['results_storage']):
            self.results_storage = pathlib.Path(new_params['output_dir']+"/"+new_params['results_storage'])
        self.reset_record_files(new_params['n_temporal_units_to_sim'], new_params['register_stocks'])

    def reset_record_files(self, n_temporal_units:int, reg_stocks: bool) -> None:
        """Reset results memmaps

        This method creates/resets the :class:`memmaps <numpy.memmap>` arrays used to track
        production, demand, overproduction, etc.

        Parameters
        ----------
        n_steps : int
            number of steps of the simulation (memmaps size are predefined)
        reg_stocks : bool
            If true, create/reset the stock memmap (which can be huge)

        """
        self.production_evolution = np.memmap(self.results_storage/"iotable_XVA_record", dtype='float64', mode="w+", shape=(n_temporal_units, self.n_sectors*self.n_regions))
        self.production_evolution.fill(np.nan)
        self.production_cap_evolution = np.memmap(self.results_storage/"iotable_X_max_record", dtype='float64', mode="w+", shape=(n_temporal_units, self.n_sectors*self.n_regions))
        self.production_cap_evolution.fill(np.nan)
        self.io_demand_evolution = np.memmap(self.results_storage/"io_demand_record", dtype='float64', mode="w+", shape=(n_temporal_units, self.n_sectors*self.n_regions))
        self.io_demand_evolution.fill(np.nan)
        self.final_demand_evolution = np.memmap(self.results_storage/"final_demand_record", dtype='float64', mode="w+", shape=(n_temporal_units, self.n_sectors*self.n_regions))
        self.final_demand_evolution.fill(np.nan)
        self.rebuild_demand_evolution = np.memmap(self.results_storage/"rebuild_demand_record", dtype='float64', mode="w+", shape=(n_temporal_units, self.n_sectors*self.n_regions))
        self.rebuild_demand_evolution.fill(np.nan)
        self.overproduction_evolution = np.memmap(self.results_storage/"overprodvector_record", dtype='float64', mode="w+", shape=(n_temporal_units, self.n_sectors*self.n_regions))
        self.overproduction_evolution.fill(np.nan)
        self.final_demand_unmet_evolution = np.memmap(self.results_storage/"final_demand_unmet_record", dtype='float64', mode="w+", shape=(n_temporal_units, self.n_sectors*self.n_regions))
        self.final_demand_unmet_evolution.fill(np.nan)
        self.rebuild_production_evolution = np.memmap(self.results_storage/"rebuild_prod_record", dtype='float64', mode="w+", shape=(n_temporal_units, self.n_sectors*self.n_regions))
        self.rebuild_production_evolution.fill(np.nan)
        if reg_stocks:
            self.stocks_evolution = np.memmap(self.results_storage/"stocks_record", dtype='float64', mode="w+", shape=(n_temporal_units, self.n_sectors, self.n_sectors*self.n_regions))
            self.stocks_evolution.fill(np.nan)
        self.limiting_stocks_evolution = np.memmap(self.results_storage/"limiting_stocks_record", dtype='byte', mode="w+", shape=(n_temporal_units, self.n_sectors, self.n_sectors*self.n_regions))
        self.limiting_stocks_evolution.fill(-1)


    def write_production(self, current_temporal_unit:int) -> None:
        self.production_evolution[current_temporal_unit] = self.production

    def write_production_max(self, current_temporal_unit:int) -> None:
        self.production_cap_evolution[current_temporal_unit] = self.production_cap

    def write_io_demand(self, current_temporal_unit:int) -> None:
         self.io_demand_evolution[current_temporal_unit] = self.matrix_orders.sum(axis=1)

    def write_final_demand(self, current_temporal_unit:int) -> None:
         self.final_demand_evolution[current_temporal_unit] = self.final_demand.sum(axis=1)

    def write_rebuild_demand(self, current_temporal_unit:int) -> None:
        to_write = np.full(self.n_regions*self.n_sectors,0.0)
        if (r_dem := self.tot_rebuild_demand) is not None:
            self.rebuild_demand_evolution[current_temporal_unit] = r_dem
        else:
            self.rebuild_demand_evolution[current_temporal_unit] = to_write

    def write_rebuild_prod(self, current_temporal_unit:int, rebuild_prod_agg:np.ndarray) -> None:
        self.rebuild_production_evolution[current_temporal_unit] = rebuild_prod_agg

    def write_overproduction(self, current_temporal_unit:int) -> None:
        self.overproduction_evolution[current_temporal_unit] = self.overprod

    def write_final_demand_unmet(self, current_temporal_unit:int, final_demand_unmet:np.ndarray) -> None:
        self.final_demand_unmet_evolution[current_temporal_unit] = final_demand_unmet

    def write_stocks(self, current_temporal_unit:int) -> None:
        self.stocks_evolution[current_temporal_unit] = self.matrix_stock

    def write_limiting_stocks(self, current_temporal_unit:int,
                              limiting_stock:np.ndarray) -> None:
        self.limiting_stocks_evolution[current_temporal_unit] = limiting_stock

    def write_index(self, index_file:str|pathlib.Path) -> None:
        """Write the indexes of the different dataframes of the model in a json file.

        In order to easily rebuild the dataframes from the 'raw' data, this
        method create a JSON file with all columns and indexes names, namely :

        * regions names
        * sectors names
        * final demand categories
        * number of regions, sectors and industries (regions * sectors)

        Parameters
        ----------
        index_file : pathlib.Path
            Path to the file to save the indexes.
        """

        indexes= {
            "regions":list(self.regions),
            "sectors":list(self.sectors),
            "fd_cat":list(self.fd_cat),
            "n_sectors":self.n_sectors,
            "n_regions":self.n_regions,
            "n_industries":self.n_sectors*self.n_regions
        }
        if isinstance(index_file,str):
            index_file=pathlib.Path(index_file)
        with index_file.open('w') as f:
            json.dump(indexes,f)

    def change_inv_duration(self, new_dur:int, old_dur:Optional[int]=None) -> None:
        if old_dur is None:
            old_dur = self.main_inv_dur
        old_dur = float(old_dur) / self.n_temporal_units_by_step
        new_dur = float(new_dur) / self.n_temporal_units_by_step
        logger.info("Changing (main) inventories duration from {} steps to {} steps (there are {} temporal units by step so duration is {})".format(old_dur, new_dur, self.n_temporal_units_by_step, new_dur*self.n_temporal_units_by_step))
        self.inv_duration = np.where(self.inv_duration==old_dur, new_dur, self.inv_duration)



"""
This type stub file was generated by pyright.
"""

from typing import Any, TypeVar, overload
from numpy import ndarray
from pandas._typing import Axis

"""
For compatibility with numpy libraries, pandas functions or methods have to
accept '*args' and '**kwargs' parameters to accommodate numpy arguments that
are not actually used or respected in the pandas implementation.

To ensure that users do not abuse these parameters, validation is performed in
'validators.py' to make sure that any extra parameters passed correspond ONLY
to those in the numpy signature. Part of that validation includes whether or
not the user attempted to pass in non-default values for these extraneous
parameters. As we want to discourage users from relying on these parameters
when calling the pandas implementation, we want them only to pass in the
default values for these parameters.

This module provides a set of commonly used default arguments for functions and
methods that are spread throughout the codebase. This module will make it
easier to adjust to future upstream changes in the analogous numpy signatures.
"""
AxisNoneT = TypeVar("AxisNoneT", Axis, None)
class CompatValidator:
    def __init__(self, defaults, fname=..., method: str | None = ..., max_fname_arg_count=...) -> None:
        ...
    
    def __call__(self, args, kwargs, fname=..., max_fname_arg_count=..., method: str | None = ...) -> None:
        ...
    


ARGMINMAX_DEFAULTS = ...
validate_argmin = ...
validate_argmax = ...
def process_skipna(skipna: bool | ndarray | None, args) -> tuple[bool, Any]:
    ...

def validate_argmin_with_skipna(skipna: bool | ndarray | None, args, kwargs) -> bool:
    """
    If 'Series.argmin' is called via the 'numpy' library, the third parameter
    in its signature is 'out', which takes either an ndarray or 'None', so
    check if the 'skipna' parameter is either an instance of ndarray or is
    None, since 'skipna' itself should be a boolean
    """
    ...

def validate_argmax_with_skipna(skipna: bool | ndarray | None, args, kwargs) -> bool:
    """
    If 'Series.argmax' is called via the 'numpy' library, the third parameter
    in its signature is 'out', which takes either an ndarray or 'None', so
    check if the 'skipna' parameter is either an instance of ndarray or is
    None, since 'skipna' itself should be a boolean
    """
    ...

ARGSORT_DEFAULTS: dict[str, int | str | None] = ...
validate_argsort = ...
ARGSORT_DEFAULTS_KIND: dict[str, int | None] = ...
validate_argsort_kind = ...
def validate_argsort_with_ascending(ascending: bool | int | None, args, kwargs) -> bool:
    """
    If 'Categorical.argsort' is called via the 'numpy' library, the first
    parameter in its signature is 'axis', which takes either an integer or
    'None', so check if the 'ascending' parameter has either integer type or is
    None, since 'ascending' itself should be a boolean
    """
    ...

CLIP_DEFAULTS: dict[str, Any] = ...
validate_clip = ...
@overload
def validate_clip_with_axis(axis: ndarray, args, kwargs) -> None:
    ...

@overload
def validate_clip_with_axis(axis: AxisNoneT, args, kwargs) -> AxisNoneT:
    ...

def validate_clip_with_axis(axis: ndarray | AxisNoneT, args, kwargs) -> AxisNoneT | None:
    """
    If 'NDFrame.clip' is called via the numpy library, the third parameter in
    its signature is 'out', which can takes an ndarray, so check if the 'axis'
    parameter is an instance of ndarray, since 'axis' itself should either be
    an integer or None
    """
    ...

CUM_FUNC_DEFAULTS: dict[str, Any] = ...
validate_cum_func = ...
validate_cumsum = ...
def validate_cum_func_with_skipna(skipna, args, kwargs, name) -> bool:
    """
    If this function is called via the 'numpy' library, the third parameter in
    its signature is 'dtype', which takes either a 'numpy' dtype or 'None', so
    check if the 'skipna' parameter is a boolean or not
    """
    ...

ALLANY_DEFAULTS: dict[str, bool | None] = ...
validate_all = ...
validate_any = ...
LOGICAL_FUNC_DEFAULTS = ...
validate_logical_func = ...
MINMAX_DEFAULTS = ...
validate_min = ...
validate_max = ...
RESHAPE_DEFAULTS: dict[str, str] = ...
validate_reshape = ...
REPEAT_DEFAULTS: dict[str, Any] = ...
validate_repeat = ...
ROUND_DEFAULTS: dict[str, Any] = ...
validate_round = ...
SORT_DEFAULTS: dict[str, int | str | None] = ...
validate_sort = ...
STAT_FUNC_DEFAULTS: dict[str, Any | None] = ...
SUM_DEFAULTS = ...
PROD_DEFAULTS = ...
MEDIAN_DEFAULTS = ...
validate_stat_func = ...
validate_sum = ...
validate_prod = ...
validate_mean = ...
validate_median = ...
STAT_DDOF_FUNC_DEFAULTS: dict[str, bool | None] = ...
validate_stat_ddof_func = ...
TAKE_DEFAULTS: dict[str, str | None] = ...
validate_take = ...
def validate_take_with_convert(convert: ndarray | bool | None, args, kwargs) -> bool:
    """
    If this function is called via the 'numpy' library, the third parameter in
    its signature is 'axis', which takes either an ndarray or 'None', so check
    if the 'convert' parameter is either an instance of ndarray or is None
    """
    ...

TRANSPOSE_DEFAULTS = ...
validate_transpose = ...
def validate_window_func(name, args, kwargs) -> None:
    ...

def validate_rolling_func(name, args, kwargs) -> None:
    ...

def validate_expanding_func(name, args, kwargs) -> None:
    ...

def validate_groupby_func(name, args, kwargs, allowed=...) -> None:
    """
    'args' and 'kwargs' should be empty, except for allowed kwargs because all
    of their necessary parameters are explicitly listed in the function
    signature
    """
    ...

RESAMPLER_NUMPY_OPS = ...
def validate_resampler_func(method: str, args, kwargs) -> None:
    """
    'args' and 'kwargs' should be empty because all of their necessary
    parameters are explicitly listed in the function signature
    """
    ...

def validate_minmax_axis(axis: int | None, ndim: int = ...) -> None:
    """
    Ensure that the axis argument passed to min, max, argmin, or argmax is zero
    or None, as otherwise it will be incorrectly ignored.

    Parameters
    ----------
    axis : int or None
    ndim : int, default 1

    Raises
    ------
    ValueError
    """
    ...




"""
This type stub file was generated by pyright.
"""

from typing import Callable

"""Common utilities for Numba operations"""
GLOBAL_USE_NUMBA: bool = ...
def maybe_use_numba(engine: str | None) -> bool:
    """Signal whether to use numba routines."""
    ...

def set_use_numba(enable: bool = ...) -> None:
    ...

def get_jit_arguments(engine_kwargs: dict[str, bool] | None = ..., kwargs: dict | None = ...) -> dict[str, bool]:
    """
    Return arguments to pass to numba.JIT, falling back on pandas default JIT settings.

    Parameters
    ----------
    engine_kwargs : dict, default None
        user passed keyword arguments for numba.JIT
    kwargs : dict, default None
        user passed keyword arguments to pass into the JITed function

    Returns
    -------
    dict[str, bool]
        nopython, nogil, parallel

    Raises
    ------
    NumbaUtilError
    """
    ...

def jit_user_function(func: Callable, nopython: bool, nogil: bool, parallel: bool) -> Callable:
    """
    JIT the user's function given the configurable arguments.

    Parameters
    ----------
    func : function
        user defined function
    nopython : bool
        nopython parameter for numba.JIT
    nogil : bool
        nogil parameter for numba.JIT
    parallel : bool
        parallel parameter for numba.JIT

    Returns
    -------
    function
        Numba JITed function
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from pandas._libs import missing as libmissing

"""
Ops for masked arrays.
"""
def kleene_or(left: bool | np.ndarray | libmissing.NAType, right: bool | np.ndarray | libmissing.NAType, left_mask: np.ndarray | None, right_mask: np.ndarray | None): # -> tuple[ndarray[Unknown, Unknown] | NDArray[bool_] | Any, NDArray[bool_] | Any | ndarray[Unknown, Unknown]]:
    """
    Boolean ``or`` using Kleene logic.

    Values are NA where we have ``NA | NA`` or ``NA | False``.
    ``NA | True`` is considered True.

    Parameters
    ----------
    left, right : ndarray, NA, or bool
        The values of the array.
    left_mask, right_mask : ndarray, optional
        The masks. Only one of these may be None, which implies that
        the associated `left` or `right` value is a scalar.

    Returns
    -------
    result, mask: ndarray[bool]
        The result of the logical or, and the new mask.
    """
    ...

def kleene_xor(left: bool | np.ndarray | libmissing.NAType, right: bool | np.ndarray | libmissing.NAType, left_mask: np.ndarray | None, right_mask: np.ndarray | None): # -> tuple[ndarray[Unknown, Unknown] | NDArray[bool_] | Any, ndarray[Unknown, Unknown] | NDArray[bool_]]:
    """
    Boolean ``xor`` using Kleene logic.

    This is the same as ``or``, with the following adjustments

    * True, True -> False
    * True, NA   -> NA

    Parameters
    ----------
    left, right : ndarray, NA, or bool
        The values of the array.
    left_mask, right_mask : ndarray, optional
        The masks. Only one of these may be None, which implies that
        the associated `left` or `right` value is a scalar.

    Returns
    -------
    result, mask: ndarray[bool]
        The result of the logical xor, and the new mask.
    """
    ...

def kleene_and(left: bool | libmissing.NAType | np.ndarray, right: bool | libmissing.NAType | np.ndarray, left_mask: np.ndarray | None, right_mask: np.ndarray | None): # -> tuple[ndarray[Unknown, Unknown] | NDArray[bool_] | Any, NDArray[bool_] | ndarray[Unknown, Unknown] | Any]:
    """
    Boolean ``and`` using Kleene logic.

    Values are ``NA`` for ``NA & NA`` or ``True & NA``.

    Parameters
    ----------
    left, right : ndarray, NA, or bool
        The values of the array.
    left_mask, right_mask : ndarray, optional
        The masks. Only one of these may be None, which implies that
        the associated `left` or `right` value is a scalar.

    Returns
    -------
    result, mask: ndarray[bool]
        The result of the logical xor, and the new mask.
    """
    ...

def raise_for_nan(value, method: str) -> None:
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
import numpy.typing as npt
from typing import Any, Generic, TypeVar, overload
from pandas._typing import IntervalClosedType, Timedelta, Timestamp

VALID_CLOSED: frozenset[str]
_OrderableScalarT = TypeVar("_OrderableScalarT", int, float)
_OrderableTimesT = TypeVar("_OrderableTimesT", Timestamp, Timedelta)
_OrderableT = TypeVar("_OrderableT", int, float, Timestamp, Timedelta)
class _LengthDescriptor:
    @overload
    def __get__(self, instance: Interval[_OrderableScalarT], owner: Any) -> _OrderableScalarT:
        ...
    
    @overload
    def __get__(self, instance: Interval[_OrderableTimesT], owner: Any) -> Timedelta:
        ...
    


class _MidDescriptor:
    @overload
    def __get__(self, instance: Interval[_OrderableScalarT], owner: Any) -> float:
        ...
    
    @overload
    def __get__(self, instance: Interval[_OrderableTimesT], owner: Any) -> _OrderableTimesT:
        ...
    


class IntervalMixin:
    @property
    def closed_left(self) -> bool:
        ...
    
    @property
    def closed_right(self) -> bool:
        ...
    
    @property
    def open_left(self) -> bool:
        ...
    
    @property
    def open_right(self) -> bool:
        ...
    
    @property
    def is_empty(self) -> bool:
        ...
    


class Interval(IntervalMixin, Generic[_OrderableT]):
    @property
    def left(self: Interval[_OrderableT]) -> _OrderableT:
        ...
    
    @property
    def right(self: Interval[_OrderableT]) -> _OrderableT:
        ...
    
    @property
    def closed(self) -> IntervalClosedType:
        ...
    
    mid: _MidDescriptor
    length: _LengthDescriptor
    def __init__(self, left: _OrderableT, right: _OrderableT, closed: IntervalClosedType = ...) -> None:
        ...
    
    def __hash__(self) -> int:
        ...
    
    @overload
    def __contains__(self: Interval[Timedelta], key: Timedelta | Interval[Timedelta]) -> bool:
        ...
    
    @overload
    def __contains__(self: Interval[Timestamp], key: Timestamp | Interval[Timestamp]) -> bool:
        ...
    
    @overload
    def __contains__(self: Interval[_OrderableScalarT], key: _OrderableScalarT | Interval[_OrderableScalarT]) -> bool:
        ...
    
    @overload
    def __add__(self: Interval[_OrderableTimesT], y: Timedelta) -> Interval[_OrderableTimesT]:
        ...
    
    @overload
    def __add__(self: Interval[int], y: _OrderableScalarT) -> Interval[_OrderableScalarT]:
        ...
    
    @overload
    def __add__(self: Interval[float], y: float) -> Interval[float]:
        ...
    
    @overload
    def __radd__(self: Interval[_OrderableTimesT], y: Timedelta) -> Interval[_OrderableTimesT]:
        ...
    
    @overload
    def __radd__(self: Interval[int], y: _OrderableScalarT) -> Interval[_OrderableScalarT]:
        ...
    
    @overload
    def __radd__(self: Interval[float], y: float) -> Interval[float]:
        ...
    
    @overload
    def __sub__(self: Interval[_OrderableTimesT], y: Timedelta) -> Interval[_OrderableTimesT]:
        ...
    
    @overload
    def __sub__(self: Interval[int], y: _OrderableScalarT) -> Interval[_OrderableScalarT]:
        ...
    
    @overload
    def __sub__(self: Interval[float], y: float) -> Interval[float]:
        ...
    
    @overload
    def __rsub__(self: Interval[_OrderableTimesT], y: Timedelta) -> Interval[_OrderableTimesT]:
        ...
    
    @overload
    def __rsub__(self: Interval[int], y: _OrderableScalarT) -> Interval[_OrderableScalarT]:
        ...
    
    @overload
    def __rsub__(self: Interval[float], y: float) -> Interval[float]:
        ...
    
    @overload
    def __mul__(self: Interval[int], y: _OrderableScalarT) -> Interval[_OrderableScalarT]:
        ...
    
    @overload
    def __mul__(self: Interval[float], y: float) -> Interval[float]:
        ...
    
    @overload
    def __rmul__(self: Interval[int], y: _OrderableScalarT) -> Interval[_OrderableScalarT]:
        ...
    
    @overload
    def __rmul__(self: Interval[float], y: float) -> Interval[float]:
        ...
    
    @overload
    def __truediv__(self: Interval[int], y: _OrderableScalarT) -> Interval[_OrderableScalarT]:
        ...
    
    @overload
    def __truediv__(self: Interval[float], y: float) -> Interval[float]:
        ...
    
    @overload
    def __floordiv__(self: Interval[int], y: _OrderableScalarT) -> Interval[_OrderableScalarT]:
        ...
    
    @overload
    def __floordiv__(self: Interval[float], y: float) -> Interval[float]:
        ...
    
    def overlaps(self: Interval[_OrderableT], other: Interval[_OrderableT]) -> bool:
        ...
    


def intervals_to_interval_bounds(intervals: np.ndarray, validate_closed: bool = ...) -> tuple[np.ndarray, np.ndarray, str]:
    ...

class IntervalTree(IntervalMixin):
    def __init__(self, left: np.ndarray, right: np.ndarray, closed: IntervalClosedType = ..., leaf_size: int = ...) -> None:
        ...
    
    @property
    def mid(self) -> np.ndarray:
        ...
    
    @property
    def length(self) -> np.ndarray:
        ...
    
    def get_indexer(self, target) -> npt.NDArray[np.intp]:
        ...
    
    def get_indexer_non_unique(self, target) -> tuple[npt.NDArray[np.intp], npt.NDArray[np.intp]]:
        ...
    
    _na_count: int
    @property
    def is_overlapping(self) -> bool:
        ...
    
    @property
    def is_monotonic_increasing(self) -> bool:
        ...
    
    def clear_mapping(self) -> None:
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
import pyarrow
from typing import TYPE_CHECKING
from pandas._typing import type_t
from pandas.core.dtypes.dtypes import register_extension_dtype
from pandas.core.arrays.masked import BaseMaskedArray, BaseMaskedDtype

if TYPE_CHECKING:
    ...
@register_extension_dtype
class BooleanDtype(BaseMaskedDtype):
    """
    Extension dtype for boolean data.

    .. versionadded:: 1.0.0

    .. warning::

       BooleanDtype is considered experimental. The implementation and
       parts of the API may change without warning.

    Attributes
    ----------
    None

    Methods
    -------
    None

    Examples
    --------
    >>> pd.BooleanDtype()
    BooleanDtype
    """
    name = ...
    @property
    def type(self) -> type:
        ...
    
    @property
    def kind(self) -> str:
        ...
    
    @property
    def numpy_dtype(self) -> np.dtype:
        ...
    
    @classmethod
    def construct_array_type(cls) -> type_t[BooleanArray]:
        """
        Return the array type associated with this dtype.

        Returns
        -------
        type
        """
        ...
    
    def __repr__(self) -> str:
        ...
    
    def __from_arrow__(self, array: pyarrow.Array | pyarrow.ChunkedArray) -> BooleanArray:
        """
        Construct BooleanArray from pyarrow Array/ChunkedArray.
        """
        ...
    


def coerce_to_array(values, mask=..., copy: bool = ...) -> tuple[np.ndarray, np.ndarray]:
    """
    Coerce the input values array to numpy arrays with a mask.

    Parameters
    ----------
    values : 1D list-like
    mask : bool 1D array, optional
    copy : bool, default False
        if True, copy the input

    Returns
    -------
    tuple of (values, mask)
    """
    ...

class BooleanArray(BaseMaskedArray):
    """
    Array of boolean (True/False) data with missing values.

    This is a pandas Extension array for boolean data, under the hood
    represented by 2 numpy arrays: a boolean array with the data and
    a boolean array with the mask (True indicating missing).

    BooleanArray implements Kleene logic (sometimes called three-value
    logic) for logical operations. See :ref:`boolean.kleene` for more.

    To construct an BooleanArray from generic array-like input, use
    :func:`pandas.array` specifying ``dtype="boolean"`` (see examples
    below).

    .. versionadded:: 1.0.0

    .. warning::

       BooleanArray is considered experimental. The implementation and
       parts of the API may change without warning.

    Parameters
    ----------
    values : numpy.ndarray
        A 1-d boolean-dtype array with the data.
    mask : numpy.ndarray
        A 1-d boolean-dtype array indicating missing values (True
        indicates missing).
    copy : bool, default False
        Whether to copy the `values` and `mask` arrays.

    Attributes
    ----------
    None

    Methods
    -------
    None

    Returns
    -------
    BooleanArray

    Examples
    --------
    Create an BooleanArray with :func:`pandas.array`:

    >>> pd.array([True, False, None], dtype="boolean")
    <BooleanArray>
    [True, False, <NA>]
    Length: 3, dtype: boolean
    """
    _internal_fill_value = ...
    _truthy_value = ...
    _falsey_value = ...
    _TRUE_VALUES = ...
    _FALSE_VALUES = ...
    def __init__(self, values: np.ndarray, mask: np.ndarray, copy: bool = ...) -> None:
        ...
    
    @property
    def dtype(self) -> BooleanDtype:
        ...
    
    _HANDLED_TYPES = ...





"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Sequence, TYPE_CHECKING
from pandas._typing import ArrayLike, DtypeObj, Manager
from pandas.core.indexes.api import Index
from numpy.ma.mrecords import MaskedRecords

"""
Functions for preparing various inputs passed to the DataFrame or Series
constructors before passing them to a BlockManager.
"""
if TYPE_CHECKING:
    ...
def arrays_to_mgr(arrays, columns: Index, index, *, dtype: DtypeObj | None = ..., verify_integrity: bool = ..., typ: str | None = ..., consolidate: bool = ...) -> Manager:
    """
    Segregate Series based on type and coerce into matrices.

    Needs to handle a lot of exceptional cases.
    """
    ...

def rec_array_to_mgr(data: MaskedRecords | np.recarray | np.ndarray, index, columns, dtype: DtypeObj | None, copy: bool, typ: str) -> Manager:
    """
    Extract from a masked rec array and create the manager.
    """
    ...

def fill_masked_arrays(data: MaskedRecords, arr_columns: Index) -> list[np.ndarray]:
    """
    Convert numpy MaskedRecords to ensure mask is softened.
    """
    ...

def mgr_to_mgr(mgr, typ: str, copy: bool = ...): # -> Manager:
    """
    Convert to specific type of Manager. Does not copy if the type is already
    correct. Does not guarantee a copy otherwise. `copy` keyword only controls
    whether conversion from Block->ArrayManager copies the 1D arrays.
    """
    ...

def ndarray_to_mgr(values, index, columns, dtype: DtypeObj | None, copy: bool, typ: str) -> Manager:
    ...

def dict_to_mgr(data: dict, index, columns, *, dtype: DtypeObj | None = ..., typ: str = ..., copy: bool = ...) -> Manager:
    """
    Segregate Series based on type and coerce into matrices.
    Needs to handle a lot of exceptional cases.

    Used in DataFrame.__init__
    """
    ...

def nested_data_to_arrays(data: Sequence, columns: Index | None, index: Index | None, dtype: DtypeObj | None) -> tuple[list[ArrayLike], Index, Index]:
    """
    Convert a single sequence of arrays to multiple arrays.
    """
    ...

def treat_as_nested(data) -> bool:
    """
    Check if we should use nested_data_to_arrays.
    """
    ...

def reorder_arrays(arrays: list[ArrayLike], arr_columns: Index, columns: Index | None, length: int) -> tuple[list[ArrayLike], Index]:
    """
    Pre-emptively (cheaply) reindex arrays with new columns.
    """
    ...

def dataclasses_to_dicts(data): # -> list[dict[str, Any]]:
    """
    Converts a list of dataclass instances to a list of dictionaries.

    Parameters
    ----------
    data : List[Type[dataclass]]

    Returns
    --------
    list_dict : List[dict]

    Examples
    --------
    >>> from dataclasses import dataclass
    >>> @dataclass
    ... class Point:
    ...     x: int
    ...     y: int

    >>> dataclasses_to_dicts([Point(1, 2), Point(2, 3)])
    [{'x': 1, 'y': 2}, {'x': 2, 'y': 3}]

    """
    ...

def to_arrays(data, columns: Index | None, dtype: DtypeObj | None = ...) -> tuple[list[ArrayLike], Index]:
    """
    Return list of arrays, columns.

    Returns
    -------
    list[ArrayLike]
        These will become columns in a DataFrame.
    Index
        This will become frame.columns.

    Notes
    -----
    Ensures that len(result_arrays) == len(result_index).
    """
    ...




"""
This type stub file was generated by pyright.
"""

from typing import TYPE_CHECKING
from pandas.io.formats.format import DataFrameFormatter

"""
Module for formatting output data in console (to string).
"""
if TYPE_CHECKING:
    ...
class StringFormatter:
    """Formatter for string representation of a dataframe."""
    def __init__(self, fmt: DataFrameFormatter, line_width: int | None = ...) -> None:
        ...
    
    def to_string(self) -> str:
        ...
    





"""
This type stub file was generated by pyright.
"""

from pandas.core.groupby.generic import DataFrameGroupBy, NamedAgg, SeriesGroupBy
from pandas.core.groupby.groupby import GroupBy
from pandas.core.groupby.grouper import Grouper

__all__ = ["DataFrameGroupBy", "NamedAgg", "SeriesGroupBy", "GroupBy", "Grouper"]



"""
This type stub file was generated by pyright.
"""

import abc
from typing import TYPE_CHECKING

"""
Engine classes for :func:`~pandas.eval`
"""
if TYPE_CHECKING:
    ...
_ne_builtins = ...
class AbstractEngine(metaclass=abc.ABCMeta):
    """Object serving as a base class for all engines."""
    has_neg_frac = ...
    def __init__(self, expr) -> None:
        ...
    
    def convert(self) -> str:
        """
        Convert an expression for evaluation.

        Defaults to return the expression as a string.
        """
        ...
    
    def evaluate(self) -> object:
        """
        Run the engine on the expression.

        This method performs alignment which is necessary no matter what engine
        is being used, thus its implementation is in the base class.

        Returns
        -------
        object
            The result of the passed expression.
        """
        ...
    


class NumExprEngine(AbstractEngine):
    """NumExpr engine class"""
    has_neg_frac = ...


class PythonEngine(AbstractEngine):
    """
    Evaluate an expression in Python space.

    Mostly for testing purposes.
    """
    has_neg_frac = ...
    def evaluate(self):
        ...
    


ENGINES: dict[str, type[AbstractEngine]] = ...



"""
This type stub file was generated by pyright.
"""

import numpy as np
from pandas._typing import ArrayLike

def write_csv_rows(data: list[ArrayLike], data_index: np.ndarray, nlevels: int, cols: np.ndarray, writer: object) -> None:
    ...

def convert_json_to_lines(arr: str) -> str:
    ...

def max_len_string_array(arr: np.ndarray) -> int:
    ...

def word_len(val: object) -> int:
    ...

def string_array_replace_from_nan_rep(arr: np.ndarray, nan_rep: object, replace: object = ...) -> None:
    ...




"""
This type stub file was generated by pyright.
"""

ne = ...
NUMEXPR_INSTALLED = ...
if NUMEXPR_INSTALLED:
    NUMEXPR_VERSION = ...
else:
    NUMEXPR_VERSION = ...
__all__ = ["NUMEXPR_INSTALLED", "NUMEXPR_VERSION"]



"""
This type stub file was generated by pyright.
"""

import pyarrow
from typing import TYPE_CHECKING
from pandas._libs import missing as libmissing
from pandas._typing import Scalar, type_t
from pandas.core.dtypes.base import StorageExtensionDtype, register_extension_dtype
from pandas.core.arrays import ExtensionArray
from pandas.core.arrays.numpy_ import PandasArray
from pandas import Series

if TYPE_CHECKING:
    ...
@register_extension_dtype
class StringDtype(StorageExtensionDtype):
    """
    Extension dtype for string data.

    .. versionadded:: 1.0.0

    .. warning::

       StringDtype is considered experimental. The implementation and
       parts of the API may change without warning.

    Parameters
    ----------
    storage : {"python", "pyarrow"}, optional
        If not given, the value of ``pd.options.mode.string_storage``.

    Attributes
    ----------
    None

    Methods
    -------
    None

    Examples
    --------
    >>> pd.StringDtype()
    string[python]

    >>> pd.StringDtype(storage="pyarrow")
    string[pyarrow]
    """
    name = ...
    @property
    def na_value(self) -> libmissing.NAType:
        ...
    
    _metadata = ...
    def __init__(self, storage=...) -> None:
        ...
    
    @property
    def type(self) -> type[str]:
        ...
    
    @classmethod
    def construct_from_string(cls, string): # -> Self@StringDtype:
        """
        Construct a StringDtype from a string.

        Parameters
        ----------
        string : str
            The type of the name. The storage type will be taking from `string`.
            Valid options and their storage types are

            ========================== ==============================================
            string                     result storage
            ========================== ==============================================
            ``'string'``               pd.options.mode.string_storage, default python
            ``'string[python]'``       python
            ``'string[pyarrow]'``      pyarrow
            ========================== ==============================================

        Returns
        -------
        StringDtype

        Raise
        -----
        TypeError
            If the string is not a valid option.
        """
        ...
    
    def construct_array_type(self) -> type_t[BaseStringArray]:
        """
        Return the array type associated with this dtype.

        Returns
        -------
        type
        """
        ...
    
    def __from_arrow__(self, array: pyarrow.Array | pyarrow.ChunkedArray) -> BaseStringArray:
        """
        Construct StringArray from pyarrow Array/ChunkedArray.
        """
        ...
    


class BaseStringArray(ExtensionArray):
    """
    Mixin class for StringArray, ArrowStringArray.
    """
    ...


class StringArray(BaseStringArray, PandasArray):
    """
    Extension array for string data.

    .. versionadded:: 1.0.0

    .. warning::

       StringArray is considered experimental. The implementation and
       parts of the API may change without warning.

    Parameters
    ----------
    values : array-like
        The array of data.

        .. warning::

           Currently, this expects an object-dtype ndarray
           where the elements are Python strings
           or nan-likes (``None``, ``np.nan``, ``NA``).
           This may change without warning in the future. Use
           :meth:`pandas.array` with ``dtype="string"`` for a stable way of
           creating a `StringArray` from any sequence.

        .. versionchanged:: 1.5.0

           StringArray now accepts array-likes containing
           nan-likes(``None``, ``np.nan``) for the ``values`` parameter
           in addition to strings and :attr:`pandas.NA`

    copy : bool, default False
        Whether to copy the array of data.

    Attributes
    ----------
    None

    Methods
    -------
    None

    See Also
    --------
    array
        The recommended function for creating a StringArray.
    Series.str
        The string methods are available on Series backed by
        a StringArray.

    Notes
    -----
    StringArray returns a BooleanArray for comparison methods.

    Examples
    --------
    >>> pd.array(['This is', 'some text', None, 'data.'], dtype="string")
    <StringArray>
    ['This is', 'some text', <NA>, 'data.']
    Length: 4, dtype: string

    Unlike arrays instantiated with ``dtype="object"``, ``StringArray``
    will convert the values to strings.

    >>> pd.array(['1', 1], dtype="object")
    <PandasArray>
    ['1', 1]
    Length: 2, dtype: object
    >>> pd.array(['1', 1], dtype="string")
    <StringArray>
    ['1', '1']
    Length: 2, dtype: string

    However, instantiating StringArrays directly with non-strings will raise an error.

    For comparison methods, `StringArray` returns a :class:`pandas.BooleanArray`:

    >>> pd.array(["a", None, "c"], dtype="string") == "a"
    <BooleanArray>
    [True, <NA>, False]
    Length: 3, dtype: boolean
    """
    _typ = ...
    def __init__(self, values, copy=...) -> None:
        ...
    
    def __arrow_array__(self, type=...):
        """
        Convert myself into a pyarrow Array.
        """
        ...
    
    def __setitem__(self, key, value): # -> None:
        ...
    
    def astype(self, dtype, copy: bool = ...): # -> Self@StringArray | IntegerArray | FloatingArray | ExtensionArray | NDArray[Unknown] | ndarray[Unknown, Unknown]:
        ...
    
    def min(self, axis=..., skipna: bool = ..., **kwargs) -> Scalar:
        ...
    
    def max(self, axis=..., skipna: bool = ..., **kwargs) -> Scalar:
        ...
    
    def value_counts(self, dropna: bool = ...) -> Series:
        ...
    
    def memory_usage(self, deep: bool = ...) -> int:
        ...
    
    _arith_method = ...
    _str_na_value = ...





"""
This type stub file was generated by pyright.
"""

from enum import Enum

_attrname_to_abbrevs: dict[str, str]
_period_code_map: dict[str, int]
def periods_per_day(reso: int) -> int:
    ...

def periods_per_second(reso: int) -> int:
    ...

def is_supported_unit(reso: int) -> bool:
    ...

def npy_unit_to_abbrev(reso: int) -> str:
    ...

class PeriodDtypeBase:
    _dtype_code: int
    def __new__(cls, code: int):
        ...
    


class FreqGroup(Enum):
    FR_ANN: int
    FR_QTR: int
    FR_MTH: int
    FR_WK: int
    FR_BUS: int
    FR_DAY: int
    FR_HR: int
    FR_MIN: int
    FR_SEC: int
    FR_MS: int
    FR_US: int
    FR_NS: int
    FR_UND: int
    @staticmethod
    def from_period_dtype_code(code: int) -> FreqGroup:
        ...
    


class Resolution(Enum):
    RESO_NS: int
    RESO_US: int
    RESO_MS: int
    RESO_SEC: int
    RESO_MIN: int
    RESO_HR: int
    RESO_DAY: int
    RESO_MTH: int
    RESO_QTR: int
    RESO_YR: int
    def __lt__(self, other: Resolution) -> bool:
        ...
    
    def __ge__(self, other: Resolution) -> bool:
        ...
    
    @property
    def attrname(self) -> str:
        ...
    
    @classmethod
    def from_attrname(cls, attrname: str) -> Resolution:
        ...
    
    @classmethod
    def get_reso_from_freqstr(cls, freq: str) -> Resolution:
        ...
    
    @property
    def attr_abbrev(self) -> str:
        ...
    


class NpyDatetimeUnit(Enum):
    NPY_FR_Y: int
    NPY_FR_M: int
    NPY_FR_W: int
    NPY_FR_D: int
    NPY_FR_h: int
    NPY_FR_m: int
    NPY_FR_s: int
    NPY_FR_ms: int
    NPY_FR_us: int
    NPY_FR_ns: int
    NPY_FR_ps: int
    NPY_FR_fs: int
    NPY_FR_as: int
    NPY_FR_GENERIC: int
    ...


